from io import BytesIO
import json

from cjio import geom_help

import numpy as np

from cjio.geom_help import triangle_normal, average_normal


def flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, str):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result


def byte_offset(x, byte_boundary):
    """Compute the byteOffset for glTF bufferView

    The bufferViews need to be aligned to a 4-byte boundary so the accessors can be aligned to them
    """
    remainder = x % byte_boundary
    padding = byte_boundary - remainder if remainder > 0 else 0
    res = x + padding
    return (res, padding)


def to_b3dm(cm, glb):
    """Convert a CityJSON to batched 3d model"""
    # glb is a buffered I/O, as the output of to_gltf()
    assert isinstance(glb, BytesIO)
    b3dm_bin = BytesIO()
    if glb.tell() == 0:
        return b3dm_bin

    # -- Feature table
    # the gltf must have a batchId per CityObject, and this setup expects that there is 1 mesh.primitive per CityObject
    feature_table_header_b = json.dumps(
        {"BATCH_LENGTH": len(cm.j["CityObjects"])}
    ).encode("utf-8")
    # The JSON header must end on an 8-byte boundary within the containing tile binary. The JSON header must be padded with trailing Space characters (0x20) to satisfy this requirement.
    offset, padding = byte_offset(len(feature_table_header_b), 8)
    for i in range(padding):
        feature_table_header_b += " ".encode("utf-8")

    # -- Batch table
    attributes = set()
    for coi_node_idxd, co in cm.j["CityObjects"].items():
        try:
            attributes |= set(co["attributes"].keys())
        except KeyError:
            pass
    if len(attributes) > 0:
        batch_table = {attribute: [] for attribute in attributes}
        for coi_node_idxd, co in cm.j["CityObjects"].items():
            for attribute in attributes:
                try:
                    batch_table[attribute].append(co["attributes"][attribute])
                except KeyError:
                    batch_table[attribute].append(None)
    else:
        batch_table = dict()
    # TODO B: write the attributes into the binary body of the the batch table
    batch_table_header_b = json.dumps(batch_table).encode("utf-8")
    offset, padding = byte_offset(len(batch_table_header_b), 8)
    for i in range(padding):
        batch_table_header_b += " ".encode("utf-8")

    # -- binary glTF
    offset, padding = byte_offset(glb.tell(), 8)
    glb.write(bytearray(padding))

    # the b3dm header is 28-bytes
    byte_length = (
        28 + len(feature_table_header_b) + len(batch_table_header_b) + glb.tell()
    )
    # -- b3dm Header
    magic = "b3dm"
    version = 1
    feature_table_json_blen = len(feature_table_header_b)
    feature_table_bin_blen = 0
    batch_table_json_blen = len(batch_table_header_b)
    batch_table_bin_blen = 0

    b3dm_bin.write(magic.encode("utf-8"))
    b3dm_bin.write(version.to_bytes(4, byteorder="little", signed=False))
    b3dm_bin.write(byte_length.to_bytes(4, byteorder="little", signed=False))
    b3dm_bin.write(
        feature_table_json_blen.to_bytes(4, byteorder="little", signed=False)
    )
    b3dm_bin.write(feature_table_bin_blen.to_bytes(4, byteorder="little", signed=False))
    b3dm_bin.write(batch_table_json_blen.to_bytes(4, byteorder="little", signed=False))
    b3dm_bin.write(batch_table_bin_blen.to_bytes(4, byteorder="little", signed=False))
    b3dm_bin.write(feature_table_header_b)
    b3dm_bin.write(batch_table_header_b)
    b3dm_bin.write(glb.getvalue())

    assert b3dm_bin.tell() == byte_length

    return b3dm_bin


def to_glb(cm, do_triangulate=True):
    """Convert to Binary glTF (.glb)

    Adapted from CityJSON2glTF: https://github.com/tudelft3d/CityJSON2glTF
    """
    gltf_json = {
        "asset": {},
        "buffers": [],
        "bufferViews": [],
        "accessors": [],
        "materials": [],
        "meshes": [],
        "nodes": [],
        "scenes": [],
    }
    gltf_bin = bytearray()
    glb = BytesIO()
    try:
        if len(cm.j["CityObjects"]) == 0:
            return glb
    except KeyError:
        raise TypeError("Not a CityJSON")

    # asset
    asset = dict()
    asset["copyright"] = "Open data"
    asset["generator"] = "Generated using cjio's glTF exporter"
    asset["version"] = "2.0"
    gltf_json["asset"] = asset

    # index bufferview
    bufferViews = []
    meshes = []
    poscount = 0
    nodes = []
    child_node_indices = []
    accessors = []
    matid = 0
    material_ids = []

    vertexlist = np.array(cm.j["vertices"])

    # gltf uses a right-handed coordinate system.
    # glTF defines +Y as up, +Z as forward, and -X as right, thus the front of a glTF
    # asset faces +Z.
    # The root node of the scene contains the y-up to z-up transformation matrix,
    # which is needed for inherently z-up data.
    # See https://github.com/CesiumGS/3d-tiles/tree/main/specification#y-up-to-z-up
    name = (
        cm.j.get("metadata", "citymodel").get("identifier", "citymodel")
        if "metadata" in cm.j
        else ""
    )
    nodes.append(
        {
            "matrix": [1, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 1],
            "children": [],
            "name": name,
        }
    )
    root_node_idx = 0

    # CityObject index with geometry that goes into the glb and is represented by a
    # mesh.
    mesh_idx = 0
    # Since we have an empty root node (without a mesh), but then we have a flat
    # hierarchy (single level) under the root node, each mesh is represented by one
    # (child) node. Technically, we could use 'mesh_idx + 1' in the code below as the
    # child node index, but we use a separate variable for extra clarity.
    child_node_idx = 1

    for theid in cm.j["CityObjects"]:
        forimax = []
        normals_per_geom = []

        if (
            "geometry" in cm.j["CityObjects"][theid]
            and len(cm.j["CityObjects"][theid]["geometry"]) != 0
        ):
            comType = cm.j["CityObjects"][theid]["type"]
            if (
                comType == "Building"
                or comType == "BuildingPart"
                or comType == "BuildingInstallation"
            ):
                matid = 0
            elif comType == "TINRelief":
                matid = 1
            elif (
                comType == "Road"
                or comType == "Railway"
                or comType == "TransportSquare"
            ):
                matid = 2
            elif comType == "WaterBody":
                matid = 3
            elif comType == "PlantCover" or comType == "SolitaryVegetationObject":
                matid = 4
            elif comType == "LandUse":
                matid = 5
            elif comType == "CityFurniture":
                matid = 6
            elif (
                comType == "Bridge"
                or comType == "BridgePart"
                or comType == "BridgeInstallation"
                or comType == "BridgeConstructionElement"
            ):
                matid = 7
            elif (
                comType == "Tunnel"
                or comType == "TunnelPart"
                or comType == "TunnelInstallation"
            ):
                matid = 8
            elif comType == "GenericCityObject":
                matid = 9
            material_ids.append(matid)

            # ----- computing vertex normals
            # We are computing soft-shading, which means computing the weighted average
            # normal for each vertex.
            # http://wiki.polycount.com/w/images/e/e5/FrostSoft_doc-3.png

            if do_triangulate:
                for geom in cm.j["CityObjects"][theid]["geometry"]:
                    normals_per_vertex = {i: [] for i in range(len(vertexlist))}
                    poscount = poscount + 1
                    if geom["type"] == "Solid":
                        triList = []
                        for shell in geom["boundaries"]:
                            for face in shell:
                                tri, success = geom_help.triangulate_face(
                                    face, vertexlist
                                )
                                if success:
                                    for t in tri:
                                        triList.append(list(t))
                                        tri_normal = triangle_normal(
                                            t, vertexlist, weighted=True
                                        )
                                        if tri_normal is not None:
                                            # for gltf, we need to invert the vector
                                            tri_normal = tri_normal * -1.0
                                            normals_per_vertex[t[0]].append(tri_normal)
                                            normals_per_vertex[t[1]].append(tri_normal)
                                            normals_per_vertex[t[2]].append(tri_normal)
                                else:
                                    # TODO: logging
                                    print(
                                        f"Failed to triangulate face in CityObject {theid}"
                                    )
                        trigeom = flatten(triList)

                    elif (geom["type"] == "MultiSurface") or (
                        geom["type"] == "CompositeSurface"
                    ):
                        triList = []
                        for face in geom["boundaries"]:
                            tri, success = geom_help.triangulate_face(face, vertexlist)
                            if success:
                                for t in tri:
                                    triList.append(list(t))
                                    tri_normal = triangle_normal(
                                        t, vertexlist, weighted=True
                                    )
                                    if tri_normal is not None:
                                        tri_normal = tri_normal * -1.0
                                        normals_per_vertex[t[0]].append(tri_normal)
                                        normals_per_vertex[t[1]].append(tri_normal)
                                        normals_per_vertex[t[2]].append(tri_normal)
                            else:
                                # TODO: logging
                                print(
                                    f"Failed to triangulate face in CityObject {theid}"
                                )
                        trigeom = flatten(triList)

                    forimax.append(trigeom)
                    # Computing smooth-shading (smooth-normal) for a vertex, as the sum of
                    # normals weighted by the triangle area of the adjacent triangles.
                    normals_per_vertex_smooth = {
                        i: None for i in range(len(vertexlist))
                    }
                    for v, normals in normals_per_vertex.items():
                        normals_per_vertex_smooth[v] = average_normal(normals)
                    del normals_per_vertex
                    normals_per_geom.append(
                        list(normals_per_vertex_smooth[v] for v in trigeom)
                    )
            else:
                # If the caller says it's triangulate, then we trust that it's
                # triangulated.
                for geom in cm.j["CityObjects"][theid]["geometry"]:
                    normals_per_vertex = {i: [] for i in range(len(vertexlist))}
                    poscount = poscount + 1
                    if geom["type"] == "Solid":
                        triList = []
                        for shell in geom["boundaries"]:
                            for face in shell:
                                for t in face:
                                    triList.append(t)
                                    tri_normal = triangle_normal(
                                        t, vertexlist, weighted=True
                                    )
                                    if tri_normal is not None:
                                        tri_normal = tri_normal * -1.0
                                        normals_per_vertex[t[0]].append(tri_normal)
                                        normals_per_vertex[t[1]].append(tri_normal)
                                        normals_per_vertex[t[2]].append(tri_normal)

                        trigeom = flatten(triList)

                    elif (geom["type"] == "MultiSurface") or (
                        geom["type"] == "CompositeSurface"
                    ):
                        triList = []
                        for face in geom["boundaries"]:
                            for t in face:
                                triList.append(t)
                                tri_normal = triangle_normal(
                                    t, vertexlist, weighted=True
                                )
                                if tri_normal is not None:
                                    tri_normal = tri_normal * -1.0
                                    normals_per_vertex[t[0]].append(tri_normal)
                                    normals_per_vertex[t[1]].append(tri_normal)
                                    normals_per_vertex[t[2]].append(tri_normal)
                        trigeom = flatten(triList)

                    forimax.append(trigeom)
                    # Computing smooth-shading (smooth-normal) for a vertex, as the sum of
                    # normals weighted by the triangle area of the adjacent triangles.
                    normals_per_vertex_smooth = {
                        i: None for i in range(len(vertexlist))
                    }
                    for v, normals in normals_per_vertex.items():
                        normals_per_vertex_smooth[v] = average_normal(normals)
                    del normals_per_vertex
                    normals_per_geom.append(
                        list(normals_per_vertex_smooth[v] for v in trigeom)
                    )

            # Flatten the triangle-vertex lists for each CityObject geometry into a
            # single list. Each consecutive set of three vertices defines a
            # single triangle primitive.
            flatgeom = flatten(forimax)
            # Same for the normal-vertex lists
            normals_np = np.concatenate(normals_per_geom)
            if len(normals_np) != len(flatgeom):
                raise RuntimeError("The length of vertices and normals should be equal")
            del normals_per_geom

            # ----- buffer and bufferView
            # allocate for vertex coordinates
            vtx_np = np.zeros((len(flatgeom), 3))
            # allocate for vertex indices
            vtx_idx_np = np.zeros(len(flatgeom))
            # need to reindex the vertices, otherwise if the vtx index exceeds the nr. of vertices in the
            # accessor then we get "ACCESSOR_INDEX_OOB" error
            for i, v in enumerate(flatgeom):
                try:
                    vtx_np[i] = np.array(
                        (vertexlist[v][0], vertexlist[v][1], vertexlist[v][2])
                    )
                except IndexError:
                    print(i, v)
                vtx_idx_np[i] = i
            bin_vtx = vtx_np.astype(np.float32).tobytes()
            # convert geometry indices to binary
            bin_geom = vtx_idx_np.astype(np.uint32).tobytes()
            del flatgeom
            # convert the normal to binary
            bin_normals = normals_np.astype(np.float32).tobytes()
            # convert batchid to binary
            batchid_np = np.array([i for g in vtx_idx_np])
            _ = batchid_np.astype(np.uint32).tobytes()

            # -- geometry indices bufferView
            bpos = len(gltf_bin)
            offset, padding = byte_offset(bpos, 4)
            bufferView = dict()
            bufferView["buffer"] = 0
            bufferView["byteOffset"] = offset
            bufferView["byteLength"] = len(bin_geom)
            bufferView["target"] = (
                34963  # For 'target' property constants see: https://github.com/KhronosGroup/glTF-Tutorials/blob/master/gltfTutorial/gltfTutorial_005_BuffersBufferViewsAccessors.md#bufferviews
            )
            # write to the buffer
            gltf_bin.extend(bin_geom)
            gltf_bin.extend(bytearray(padding))
            bufferViews.append(bufferView)

            # -- geometry vertices (POSITION) bufferView
            bpos = len(gltf_bin)
            offset, padding = byte_offset(bpos, 4)
            bufferView = dict()
            bufferView["buffer"] = 0
            bufferView["byteOffset"] = offset
            # bufferView["byteStride"] = 12
            bufferView["byteLength"] = len(bin_vtx)
            bufferView["target"] = 34962
            # write to the buffer
            gltf_bin.extend(bin_vtx)
            gltf_bin.extend(bytearray(padding))
            bufferViews.append(bufferView)

            # -- vertex normals (NORMAL) bufferView
            bpos = len(gltf_bin)
            offset, padding = byte_offset(bpos, 4)
            bufferView = dict()
            bufferView["buffer"] = 0
            bufferView["byteOffset"] = offset
            # bufferView["byteStride"] = 12
            bufferView["byteLength"] = len(bin_normals)
            bufferView["target"] = 34962
            # write to the buffer
            gltf_bin.extend(bin_normals)
            gltf_bin.extend(bytearray(padding))
            bufferViews.append(bufferView)

            # #-- batchid bufferView
            # bpos = len(gltf_bin)
            # offset, padding = byte_offset(bpos, 4)
            # bufferView = dict()
            # bufferView["buffer"] = 0
            # bufferView["byteOffset"] = offset
            # bufferView["byteStride"] = 4
            # bufferView["byteLength"] = len(bin_batchid)
            # bufferView["target"] = 34962
            # # write to the buffer
            # gltf_bin.extend(bin_batchid)
            # gltf_bin.extend(bytearray(padding))
            # bufferViews.append(bufferView)

            # ----- accessors

            # accessor for geometry indices bufferView
            accessor = dict()
            # We start the bufferViews with i==1, because the root node does not have
            # a bufferView, because it is empty (does not have a mesh)
            # Without a root node we would need  = 0 if coi_node_idx == 0  else coi_node_idx * 3
            accessor["bufferView"] = mesh_idx * 3
            accessor["componentType"] = 5125
            accessor["count"] = len(vtx_idx_np)
            accessor["type"] = "SCALAR"
            accessor["max"] = [int(vtx_idx_np.max())]
            accessor["min"] = [int(vtx_idx_np.min())]
            accessors.append(accessor)

            # accessor for geometry vertices (POSITION) bufferView
            accessor = dict()
            # without an empty root node we would need = 1 if coi_node_idx == 0  else coi_node_idx * 3 + 1
            accessor["bufferView"] = mesh_idx * 3 + 1
            accessor["componentType"] = 5126
            accessor["count"] = len(vtx_np)
            accessor["type"] = "VEC3"
            accessor["max"] = [
                float(np.amax(vtx_np, axis=0)[0]),
                float(np.amax(vtx_np, axis=0)[1]),
                float(np.amax(vtx_np, axis=0)[2]),
            ]
            accessor["min"] = [
                float(np.amin(vtx_np, axis=0)[0]),
                float(np.amin(vtx_np, axis=0)[1]),
                float(np.amin(vtx_np, axis=0)[2]),
            ]
            accessors.append(accessor)

            # accessor for vertex normals (NORMAL) bufferView
            accessor = dict()
            accessor["bufferView"] = mesh_idx * 3 + 2
            accessor["componentType"] = 5126
            accessor["count"] = len(normals_np)
            accessor["type"] = "VEC3"
            accessor["max"] = [
                float(np.amax(normals_np, axis=0)[0]),
                float(np.amax(normals_np, axis=0)[1]),
                float(np.amax(normals_np, axis=0)[2]),
            ]
            accessor["min"] = [
                float(np.amin(normals_np, axis=0)[0]),
                float(np.amin(normals_np, axis=0)[1]),
                float(np.amin(normals_np, axis=0)[2]),
            ]
            accessors.append(accessor)

            # # accessor for batchid bufferView
            # accessor = dict()
            # # without an empty root node we would need = 2 if coi_node_idx == 0  else coi_node_idx * 3 + 2
            # accessor["bufferView"] = mesh_idx * 3 + 3
            # accessor["componentType"] = 5123
            # accessor["count"] = len(batchid_np)
            # accessor["type"] = "SCALAR"
            # accessors.append(accessor)

            # ----- meshes
            # one mesh per CityObject
            mesh = dict()
            mesh["name"] = str(theid)
            mesh["primitives"] = [
                {
                    "indices": len(accessors) - 3,
                    "material": matid,
                    "attributes": {
                        # "_BATCHID": len(accessors) - 1,
                        "NORMAL": len(accessors) - 1,
                        "POSITION": len(accessors) - 2,
                    },
                }
            ]
            meshes.append(mesh)

            # ----- nodes
            # a node has a mesh, and the mesh is referenced by its index in the meshes
            nodes.append({"mesh": mesh_idx})
            # one node per CityObject
            child_node_indices.append(child_node_idx)

            mesh_idx += 1
            child_node_idx += 1

    # -- buffers
    buffer = dict()
    offset, padding = byte_offset(len(gltf_bin), 4)
    gltf_bin.extend(bytearray(padding))
    buffer["byteLength"] = len(gltf_bin)
    gltf_json["buffers"] = [buffer]

    gltf_json["bufferViews"] = bufferViews
    gltf_json["accessors"] = accessors
    gltf_json["meshes"] = meshes

    nodes[root_node_idx]["children"] = child_node_indices
    gltf_json["nodes"] = nodes

    scene = dict()
    scene["nodes"] = [
        root_node_idx,
    ]
    gltf_json["scenes"] = [scene]

    # -- materials
    materials = [
        {  # building red
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.7200, 0.320, 0.220, 1.0],
                "metallicFactor": 0.0,
                "roughnessFactor": 1.0,
            }
        },
        {  # terrain brown
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.588, 0.403, 0.211, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # transport grey
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.631, 0.607, 0.592, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # waterbody blue
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.070, 0.949, 0.972, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # vegetation green
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.000, 1.000, 0.000, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # landuse yellow
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.909, 0.945, 0.196, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # CityFurniture orange
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.894, 0.494, 0.145, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # bridge purple
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.466, 0.094, 0.905, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # tunnel black
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.011, 0.011, 0.007, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
        {  # GenericCityObject pink
            "pbrMetallicRoughness": {
                "baseColorFactor": [0.909, 0.188, 0.827, 1.0],
                "metallicFactor": 0.5,
                "roughnessFactor": 1.0,
            }
        },
    ]
    gltf_json["materials"] = materials

    # -- Chunk 0 (JSON)
    chunk_0 = json.dumps(gltf_json).encode("utf-8")
    offset, padding = byte_offset(len(chunk_0), 4)
    for i in range(padding):
        chunk_0 += " ".encode("utf-8")
    # -- Binary glTF header
    magic = "glTF"
    version = 2
    length = 12 + 8 + len(chunk_0) + 8 + len(gltf_bin)

    # header
    glb.write(magic.encode("utf-8"))
    glb.write(version.to_bytes(4, byteorder="little", signed=False))
    glb.write(length.to_bytes(4, byteorder="little", signed=False))
    # chunk 0
    glb.write(len(chunk_0).to_bytes(4, byteorder="little", signed=False))
    glb.write("JSON".encode("utf-8"))
    glb.write(chunk_0)
    # chunk 1
    glb.write(len(gltf_bin).to_bytes(4, byteorder="little", signed=False))
    glb.write(bytearray.fromhex("42494e00"))  # == BIN + 1 empty byte for padding
    glb.write(gltf_bin)
    assert length == glb.tell()

    return glb


def faces_to_obj(faces, out, sloppy, vnp, texture_values=None, textures=None):
    """Writes a list of surfaces to an obj file. If textures are provided, the coordinate textures are also written.
    All n-gons with n >=4 are triangulated.

    :param faces: A list of surfaces, each surface is a list of list.
    :param out: An instance of io.StringIO, the file to write to.
    :param sloppy: See cjio.export_cmd.
    :param vnp: An array of vertices.
    :param texture_values: None or a list of texture coordinates of the same size as faces.
        Each "texture coordinates" is a list `[idx_texture, x1, x2, ..., xn]`. `idx_texture` is the index of the texture
        in the `textures` list (starting at 0) and `x1, x2, ..., xn` are the indices of the texture vertices
        corresponding to the vertices of the face.
    :param textures: None or a list of textures. Each texture is a dictionary that must have the key `name`.
    """
    current_material = None
    it = zip(faces, texture_values) if texture_values is not None else zip(faces)
    for face, *texture_vals in it:
        re, b = geom_help.triangulate_face(face, vnp, sloppy)
        v_t_map = None
        if texture_vals:
            texture_vals = texture_vals[0]
            mtl = texture_vals[0][0]
            if mtl != current_material and mtl is not None:
                out.write("usemtl " + textures[mtl]["name"] + "\n")
            current_material = mtl
            # Map each vertex to its texture coordinate
            if b and current_material is not None:
                v_t_map = {}
                for f, t in zip(face, texture_vals):
                    for v, vt in zip(
                        f, t[1:]
                    ):  # t[0] is the material, t[1:] are the texture coordinates
                        v_t_map[v] = vt
        if b:
            for v in re:
                if v_t_map is not None and current_material is not None:
                    out.write(
                        "f %d/%d %d/%d %d/%d\n"
                        % (
                            v[0] + 1,
                            v_t_map[v[0]] + 1,
                            v[1] + 1,
                            v_t_map[v[1]] + 1,
                            v[2] + 1,
                            v_t_map[v[2]] + 1,
                        )
                    )
                else:
                    out.write("f %d %d %d\n" % (v[0] + 1, v[1] + 1, v[2] + 1))



import copy
import glob
import json
import os.path
import sys
import warnings
from os import linesep
from pathlib import Path

import click

import cjio
from cjio import (
    cityjson,
    errors,
    utils,
    MODULE_TRIANGLE_AVAILABLE,
    MODULE_PYPROJ_AVAILABLE,
    MODULE_EARCUT_AVAILABLE,
    MODULE_CJVAL_AVAILABLE,
)
from cjio.floatEncoder import FloatEncoder

json.encoder.c_make_encoder = None
json.encoder.float = FloatEncoder


# -- https://stackoverflow.com/questions/47437472/in-python-click-how-do-i-see-help-for-subcommands-whose-parents-have-required
class PerCommandArgWantSubCmdHelp(click.Argument):
    def handle_parse_result(self, ctx, opts, args):
        # check to see if there is a --help on the command line
        if any(arg in ctx.help_option_names for arg in args):
            # if asking for help see if we are a subcommand name
            for arg in opts.values():
                if arg in ctx.command.commands:
                    # this matches a sub command name, and --help is
                    # present, let's assume the user wants help for the
                    # subcommand
                    args = [arg] + args
        return super(PerCommandArgWantSubCmdHelp, self).handle_parse_result(
            ctx, opts, args
        )


@click.group(chain=True)
@click.version_option(
    version=cjio.__version__,
    prog_name=cityjson.CITYJSON_VERSIONS_SUPPORTED[-1],
    message="cjio v%(version)s; supports CityJSON v%(prog)s",
)
@click.argument("input", cls=PerCommandArgWantSubCmdHelp)
@click.option(
    "--ignore_duplicate_keys",
    is_flag=True,
    help="Load a CityJSON file even if some City Objects have the same IDs (technically invalid file).",
)
@click.option("--suppress_msg", is_flag=True, help="Suppress all information/messages.")
@click.pass_context
def cli(context, input, ignore_duplicate_keys, suppress_msg):
    """Process and manipulate a CityJSON model, and allow
    different outputs. The different operators can be chained
    to perform several processing in one step, the CityJSON model
    goes through the different operators.

    Can read a file or from stdin (with a JSONL file as input).

    To get help on specific command, eg for 'validate':

    \b
        cjio validate --help

    Usage examples:

    \b
        cjio myfile.city.json info
        cjio myfile.city.json subset --id house12 save out.city.json
        cjio myfile.city.json crs_assign 7145 textures_remove export --format obj output.obj
        cat mystream.city.jsonl | cjio stdin info
    """
    context.ensure_object(dict)
    context.obj = {"argument": input, "suppress_msg": suppress_msg}


@cli.result_callback()
def process_pipeline(processors, input, ignore_duplicate_keys, suppress_msg):
    extensions = [".json", ".jsonl", ".off", ".poly"]  # -- input allowed
    try:
        if input == "stdin":
            cm = cityjson.read_stdin()
        else:
            f = click.open_file(input, mode="r", encoding="utf-8-sig")
            extension = os.path.splitext(input)[1].lower()
            if extension not in extensions:
                raise IOError(
                    "File type not supported (only .json, .jsonl, .off, and .poly)."
                )
            # -- OFF file
            if extension == ".off":
                print_cmd_status("Converting %s to CityJSON" % (input))
                cm = cityjson.off2cj(f)
            # -- POLY file
            elif extension == ".poly":
                print_cmd_status("Converting %s to CityJSON" % (input))
                cm = cityjson.poly2cj(f)
            # -- CityJSON file
            else:
                print_cmd_status("Parsing %s" % (input))
                cm = cityjson.reader(
                    file=f, ignore_duplicate_keys=ignore_duplicate_keys
                )
                try:
                    with warnings.catch_warnings(record=True) as w:
                        cm.check_version()
                        print_cmd_alert(w)
                except errors.CJInvalidVersion as e:
                    raise click.ClickException(e.msg)
    except ValueError as e:
        raise click.ClickException('%s: "%s".' % (e, input))
    except IOError as e:
        raise click.ClickException('Invalid file: "%s".\n%s' % (input, e))
    for processor in processors:
        cm = processor(cm)


@cli.command("print")
def print_cmd():
    """Print the (pretty formatted) JSON to the console."""

    def processor(cm):
        json_str = json.dumps(cm.j, indent="  ")
        print_cmd_info(json_str)
        return cm

    return processor


@cli.command("info")
@click.option("--long", is_flag=True, help="More details about the file.")
def info_cmd(long):
    """Output information about the dataset."""

    def processor(cm):
        print_cmd_status("=== Information ===")
        s = linesep.join(cm.get_info(long=long))
        print_cmd_info(s)
        print_cmd_status("=== Information ===")
        return cm

    return processor


@cli.command("export")
@click.argument(
    "format", type=click.Choice(["jsonl", "obj", "stl", "glb", "b3dm"]), required=True
)
@click.argument("filename")
@click.option(
    "--sloppy",
    is_flag=True,
    help="Use a more lenient triangulator (mapbox-earcut), which is also less robust.",
)
def export_cmd(filename, format, sloppy):
    """Export to another format.

    CityJSONSeq (JSONL/JSON Lines for streaming), OBJ, Binary glTF (glb), Batched 3DModel (b3dm), STL.
    The result can be stored either in a file, out piped to stdout (by choosing 'stdout' instead
    of a file).

    Currently, textures are only supported for CityJSONSeq and OBJ export.

    Usage examples:

    \b
        cjio myfile.city.json export obj myfile.obj
        cjio myfile.city.json export --sloppy obj myfile.obj
        cjio --suppress_msg myfile.city.json export jsonl stdout
    """

    def exporter(cm, sloppy):
        stdoutoutput = False
        if filename == "stdout":
            stdoutoutput = True
        else:
            output = utils.verify_filename(filename)
            if output["dir"]:
                os.makedirs(output["path"], exist_ok=True)
                input_filename = os.path.splitext(os.path.basename(cm.path))[0]
                output["path"] = os.path.join(
                    output["path"], "{f}.{ext}".format(f=input_filename, ext=format)
                )
            else:
                os.makedirs(os.path.dirname(output["path"]), exist_ok=True)
        # ---------- OBJ ----------
        if format.lower() == "obj":
            if stdoutoutput:
                buf = cm.export2obj(sloppy)
                buf.seek(0)
                for line in buf.readlines():
                    sys.stdout.write(line)
            else:
                print_cmd_status("Exporting CityJSON to OBJ (%s)" % (output["path"]))
                try:
                    mtl_path = Path(output["path"]).with_suffix(".mtl")
                    re = cm.export2obj(sloppy, mtl_path.name)
                    if isinstance(re, tuple):  # MTL file returned
                        obj, mtl = re[0], re[1]
                    else:
                        obj, mtl = re, None
                    # Write .obj
                    with click.open_file(output["path"], mode="w") as fo:
                        fo.write(obj.getvalue())
                    if mtl is not None:
                        # Write optional .mtl
                        with click.open_file(str(mtl_path), mode="w") as fmtl:
                            fmtl.write(mtl.getvalue())
                except IOError as e:
                    raise click.ClickException(
                        'Invalid output file: "%s".\n%s' % (output["path"], e)
                    )
        # ---------- STL ----------
        elif format.lower() == "stl":
            if stdoutoutput:
                buf = cm.export2stl(sloppy)
                buf.seek(0)
                for line in buf.readlines():
                    sys.stdout.write(line)
            else:
                print_cmd_status("Exporting CityJSON to STL (%s)" % (output["path"]))
                try:
                    with click.open_file(output["path"], mode="w") as fo:
                        re = cm.export2stl(sloppy)
                        fo.write(re.getvalue())
                except IOError as e:
                    raise click.ClickException(
                        'Invalid output file: "%s".\n%s' % (output["path"], e)
                    )
        # ---------- GLB ----------
        elif format.lower() == "glb":
            # -- TODO: glb stdout necessary?
            fname = os.path.splitext(os.path.basename(output["path"]))[0]
            bufferbin = "{}.glb".format(fname)
            binfile = os.path.join(os.path.dirname(output["path"]), bufferbin)
            print_cmd_status("Exporting CityJSON to glb %s" % binfile)
            glb = cm.export2glb()
            # TODO B: how many buffer can there be in the 'buffers'?
            try:
                glb.seek(0)
                with click.open_file(binfile, mode="wb") as bo:
                    bo.write(glb.getvalue())
            except IOError as e:
                raise click.ClickException(
                    'Invalid output file: "%s".\n%s' % (binfile, e)
                )
        # ---------- B3DM ----------
        elif format.lower() == "b3dm":
            # -- TODO: b3dm stdout necessary?
            fname = os.path.splitext(os.path.basename(output["path"]))[0]
            b3dmbin = "{}.b3dm".format(fname)
            binfile = os.path.join(os.path.dirname(output["path"]), b3dmbin)
            b3dm = cm.export2b3dm()
            print_cmd_status("Exporting CityJSON to b3dm %s" % binfile)
            print_cmd_warning(
                "Although the conversion works, the output is probably incorrect."
            )
            try:
                b3dm.seek(0)
                with click.open_file(binfile, mode="wb") as bo:
                    bo.write(b3dm.getvalue())
            except IOError as e:
                raise click.ClickException(
                    'Invalid output file: "%s".\n%s' % (binfile, e)
                )
        # ---------- JSONL ----------
        elif format.lower() == "jsonl":
            if stdoutoutput:
                with warnings.catch_warnings(record=True) as w:
                    buf = cm.export2jsonl()
                    print_cmd_warning(w)
                buf.seek(0)
                for line in buf.readlines():
                    sys.stdout.write(line)
            else:
                print_cmd_status(
                    "Exporting CityJSON to JSON Lines (%s)" % (output["path"])
                )
                try:
                    with click.open_file(output["path"], mode="w") as fo:
                        with warnings.catch_warnings(record=True) as w:
                            re = cm.export2jsonl()
                            print_cmd_warning(w)
                        fo.write(re.getvalue())
                except IOError as e:
                    raise click.ClickException(
                        'Invalid output file: "%s".\n%s' % (output["path"], e)
                    )

    def processor(cm):
        if (format != "jsonl") and (not MODULE_TRIANGLE_AVAILABLE):
            str = "OBJ|glTF|b3dm export skipped: Python module 'triangle' missing (to triangulate faces)"
            print_cmd_alert(str)
            str = "Install it: https://pypi.org/project/triangle/"
            print_cmd_warning(str)
            raise click.ClickException("Abort.")
        else:
            exporter(cm, sloppy)
        return cm

    return processor


@cli.command("save")
@click.argument("filename")
@click.option(
    "--indent",
    is_flag=True,
    help="Indent the file. Helpful when you want to examine the file in a text editor.",
)
@click.option(
    "--textures",
    default=None,
    type=str,
    help="Path to the new textures directory. This command copies the textures to a new location. Useful when creating an independent subset of a CityJSON file.",
)
def save_cmd(filename, indent, textures):
    """Save to a CityJSON file.

    Save to a file on disk, or 'stdout' pipes the file to the standart output.

    Usage examples:

    \b
        cjio myfile.city.json metadata_update save myfile.city.json
        cjio myfile.json upgrade save stdout
    """

    def saver(cm):
        stdoutoutput = False
        if filename == "stdout":
            stdoutoutput = True
        else:
            output = utils.verify_filename(filename)
            if output["dir"]:
                os.makedirs(output["path"], exist_ok=True)
                input_filename = os.path.splitext(os.path.basename(cm.path))[0]
                output["path"] = os.path.join(
                    output["path"], "{f}.{ext}".format(f=input_filename, ext="json")
                )
            else:
                os.makedirs(os.path.dirname(output["path"]), exist_ok=True)
        if stdoutoutput:
            if indent:
                json_str = json.dumps(cm.j, indent="\t")
                sys.stdout.write(json_str)
            else:
                json_str = json.dumps(cm.j, separators=(",", ":"))
                sys.stdout.write(json_str)
        else:
            print_cmd_status("Saving CityJSON to a file %s" % output["path"])
            try:
                fo = click.open_file(output["path"], mode="w")
                if textures:
                    cm.copy_textures(textures)
                if indent:
                    json_str = json.dumps(cm.j, indent="\t")
                else:
                    json_str = json.dumps(cm.j, separators=(",", ":"))
                fo.write(json_str)
            except IOError as e:
                raise click.ClickException(
                    "Invalid output file: %s \n%s" % (output["path"], e)
                )

    def processor(cm):
        saver(cm)
        return cm

    return processor


@cli.command("validate")
def validate_cmd():
    """
    Validate the CityJSON:
    (1) against its schemas
    (2) against the (potential) Extensions schemas
    (3) extra validations

    (see https://github.com/cityjson/cjval#what-is-validated-exactly for details)

    The Extensions in the files are fetched automatically.

    \b
        $ cjio myfile.city.json validate
    """

    def processor(cm):
        if not MODULE_CJVAL_AVAILABLE:
            str = "Validation skipped: Python module 'cjvalpy' not installed"
            print_cmd_alert(str)
            str = "To install it: https://www.github.com/cityjson/cjvalpy"
            print_cmd_warning(str)
            str = "Alternatively use the web-app: https://validator.cityjson.org"
            print_cmd_warning(str)
            raise click.ClickException("Abort.")
        print_cmd_status("Validation (with official CityJSON schemas)")
        try:
            re = cm.validate()
            print_cmd_info(re)
        except Exception as e:
            print_cmd_alert("Error: {}".format(e))
        return cm

    return processor


@cli.command("merge")
@click.argument("filepattern")
def merge_cmd(filepattern):
    """
    Merge the current CityJSON with other ones.
    All City Objects with their textures/materials/templates are handled.

    Possible to give a wildcard but put it between quotes:

        $ cjio myfile.city.json merge '/home/elvis/temp/*.json' save merged.city.json
    """

    def processor(cm):
        print_cmd_status("Merging files")
        lsCMs = []
        g = glob.glob(filepattern)
        for i in g:
            try:
                f = click.open_file(i, mode="r", encoding="utf-8-sig")
                lsCMs.append(cityjson.reader(f))
            except IOError as err:
                raise click.ClickException('%s: "%s".' % (err, input))
        if len(lsCMs) == 0:
            print_cmd_info("WARNING: No files to merge.")
        else:
            cm.merge(lsCMs)
        return cm

    return processor


@cli.command("subset")
@click.option(
    "--id",
    multiple=True,
    help="The ID of the City Objects; can be used multiple times.",
)
@click.option("--bbox", nargs=4, type=float, help="2D bbox: minx miny maxx maxy.")
@click.option("--radius", nargs=3, type=float, help="x y radius")
@click.option("--random", type=int, help="Number of random City Objects to select.")
@click.option(
    "--cotype", multiple=True, help="The City Object types; can be used multiple times."
)
@click.option(
    "--exclude",
    is_flag=True,
    help="Excludes the selection, thus delete the selected object(s).",
)
def subset_cmd(id, bbox, random, cotype, radius, exclude):
    """
    Create a subset, City Objects can be selected by:
    (1) IDs of City Objects;
    (2) bbox;
    (3) City Object type(s);
    (4) randomly.

    These can be combined, except random which overwrites others.

    Option '--exclude' excludes the selected objects, or "reverse" the selection.

    Usage examples:

    \b
        cjio myfile.city.json subset --bbox 104607 490148 104703 490257 save out.city.json
        cjio myfile.city.json subset --radius 500.0 610.0 50.0 --exclude save out.city.json
        cjio myfile.city.json subset --id house12 save out.city.json
        cjio myfile.city.json subset --random 5 save out.city.json
        cjio myfile.city.json subset --cotype LandUse --cotype Building save out.city.json
    """

    def processor(cm):
        print_cmd_status("Subset of CityJSON")
        s = copy.deepcopy(cm)
        if random is not None:
            s = s.get_subset_random(random, exclude=exclude)
            return s
        elif radius is not None and len(radius) > 0:
            s = s.get_subset_radius(radius[0], radius[1], radius[2], exclude=exclude)
        elif id is not None and len(id) > 0:
            s = s.get_subset_ids(id, exclude=exclude)
        elif bbox is not None and len(bbox) > 0:
            s = s.get_subset_bbox(bbox, exclude=exclude)
        elif cotype is not None:
            s = s.get_subset_cotype(cotype, exclude=exclude)
        else:
            click.BadArgumentUsage(
                "You must provide one of the options for subset; --id, --bbox, --random, --cotype"
            )
        return s

    return processor


@cli.command("vertices_clean")
def vertices_clean_cmd():
    """
    Remove duplicate vertices + orphan vertices
    """

    def processor(cm):
        print_cmd_status("Clean the file")
        cm.remove_duplicate_vertices()
        cm.remove_orphan_vertices()
        return cm

    return processor


@cli.command("materials_remove")
def materials_remove_cmd():
    """
    Remove all materials.
    """

    def processor(cm):
        print_cmd_status("Remove all material")
        cm.remove_materials()
        return cm

    return processor


@cli.command("textures_remove")
def textures_remove_cmd():
    """
    Remove all textures.
    """

    def processor(cm):
        print_cmd_status("Remove all textures")
        cm.remove_textures()
        return cm

    return processor


@cli.command("crs_assign")
@click.argument("newepsg", type=int)
def crs_assign_cmd(newepsg):
    """
    Assign a (new) CRS (an EPSG).
    Can be used to assign one to a file that doesn't have any, or update one.

    To reproject (and thus modify all the values of the coordinates) use reproject().
    """

    def processor(cm):
        print_cmd_status("Assign EPSG:%d" % newepsg)
        cm.set_epsg(newepsg)
        return cm

    return processor


@cli.command("crs_reproject")
@click.argument("epsg", type=int)
@click.option("--digit", type=click.IntRange(1, 12), help="Number of digits to keep.")
def crs_reproject_cmd(epsg, digit):
    """
    Reproject to a new EPSG.
    The current CityJSON must have an EPSG defined
    (which can be done with function epsg_assign).
    It is possible to define the number of digits that will be kept for the result with --digit.

        $ cjio myfile.city.json crs_reproject --digit 7 4979 save newfile.city.json
    """

    def processor(cm):
        if not MODULE_PYPROJ_AVAILABLE:
            str = "Reprojection skipped: Python module 'pyproj' missing (to reproject coordinates)"
            print_cmd_alert(str)
            str = "Install it: https://pypi.org/project/pyproj/"
            print_cmd_warning(str)
            raise click.ClickException("Abort.")
        print_cmd_status("Reproject to EPSG:%d" % epsg)
        if cm.get_epsg() is None:
            print_cmd_warning(
                "WARNING: CityJSON has no EPSG defined, can't be reprojected."
            )
        else:
            with warnings.catch_warnings(record=True) as w:
                cm.reproject(epsg, digit)
                print_cmd_warning(w)
        return cm

    return processor


@cli.command("upgrade")
@click.option(
    "--digit",
    default=3,
    type=click.IntRange(1, 12),
    help="Number of digits to keep to compress.",
)
def upgrade_cmd(digit):
    """
    Upgrade the CityJSON to the latest version.
    It takes care of *everything* (touch wood).

        $ cjio myfile.city.json upgrade save upgraded.city.json

    For v1.1+, the file needs to be compressed, and you can
    speficy the number of digits to keep (default=3)

        $ cjio myfile.city.json upgrade --digit 2 save upgraded.city.json
    """

    def processor(cm):
        vlatest = cityjson.CITYJSON_VERSIONS_SUPPORTED[-1]
        print_cmd_status("Upgrade CityJSON file to v%s" % vlatest)
        re, reasons = cm.upgrade_version(vlatest, digit)
        if not re:
            print_cmd_warning("WARNING: %s" % (reasons))
        return cm

    return processor


@cli.command("textures_locate")
def textures_locate_cmd():
    """
    Print the location of the texture files.

        $ cjio input.city.json textures_locate
    """

    def processor(cm):
        print_cmd_status("Locating the textures")
        try:
            loc = cm.get_textures_location()
            if loc is None:
                print_cmd_info("This file does not have textures")
            else:
                print_cmd_info("Textures location: %s" % loc)
        except Exception as e:
            print_cmd_warning(str(e))
        return cm

    return processor


@cli.command("textures_update")
@click.argument("newlocation", type=str)
@click.option(
    "--relative", is_flag=True, help="Convert texture file paths to relative paths."
)
def textures_update_cmd(newlocation, relative):
    """
    Update the location of the texture files.
    Can be used if the texture files were moved to new directory.

      $ cjio input.city.json textures_update /new/textures/location save out.city.json
      $ cjio input.city.json textures_update /new/textures/relative/location --relative save out.city.json
    """

    def processor(cm):
        print_cmd_status("Updating location of textures")
        cm.update_textures_location(newlocation, relative=relative)
        return cm

    return processor


@cli.command("lod_filter")
@click.argument("lod", type=str)
def lod_filter_cmd(lod):
    """
    Filter only one LoD for a dataset.
    To use on datasets having more than one LoD for the city objects.
    For each city object, it keeps only the geometries having the LoD
    passed as parameter; if a city object doesn't have this LoD then
    it ends up with an empty geometry.

        $ cjio myfile.city.json lod_filter 2.2 save myfile_lod2.city.json

    """

    def processor(cm):
        print_cmd_status('Filter LoD: "%s"' % lod)
        cm.filter_lod(lod)
        return cm

    return processor


@cli.command("attribute_remove")
@click.argument("attr", type=str, nargs=1)
def attribute_remove_cmd(attr):
    """
    Remove an attribute.
    If it's not present nothing is done.
    That's it.

        $ cjio myfile.city.json attribute_remove roofType info
    """

    def processor(cm):
        print_cmd_status('Remove attribute: "%s"' % attr)
        cm.remove_attribute(attr)
        return cm

    return processor


@cli.command("attribute_rename")
@click.argument("oldattr", type=str, nargs=1)
@click.argument("newattr", type=str, nargs=1)
def attribute_rename_cmd(oldattr, newattr):
    """
    Rename an attribute.
    If it's not present nothing is done, and its value is kept.
    That's it.

        $ cjio myfile.city.json attribute_rename oldAttr newAttr info
    """

    def processor(cm):
        print_cmd_status('Rename attribute: "%s" => "%s"' % (oldattr, newattr))
        cm.rename_attribute(oldattr, newattr)
        return cm

    return processor


@cli.command("crs_translate")
@click.option("--minxyz", nargs=3, type=float, help="(x, y, z)")
def crs_translate_cmd(minxyz):
    """
    Translate the coordinates.
    By default, they are all moved by (-minx, -miny, -minz),
    so the values are smaller (often useful for further processing data).
    The CRS/EPSG is updated to 'None'.
    Three specific values for the translation can also be given with --minxyz.

        $ cjio myfile.city.json crs_translate save out.city.json
        $ cjio myfile.city.json crs_translate --minxyz -100 -25 -1 save out.city.json
    """

    def processor(cm):
        print(minxyz)
        if minxyz is None:
            bbox = cm.translate(None)
        else:
            bbox = cm.translate(minxyz)
        print_cmd_status(
            "Translating the file by: (%f, %f, %f)" % (bbox[0], bbox[1], bbox[2])
        )
        return cm

    return processor


@cli.command("metadata_get")
def metadata_get_cmd():
    """
    Show the metadata of this dataset.

    The difference between 'info' and this command is that this
    command lists the "pure" metadata as stored in the file.
    The 'info' command should be used when an overview of the
    file is needed.
    """

    def processor(cm):
        j = {}
        if cm.has_metadata():
            j.update(cm.get_metadata())
        print_cmd_info(json.dumps(j, indent=2))
        return cm

    return processor


@cli.command("metadata_extended_remove")
def metadata_remove_cmd():
    """
    Remove the deprecated +metadata-extended properties.
    Modify/update the dataset.
    """

    def processor(cm):
        print_cmd_status("Remove the +metadata-extended property")
        cm.metadata_extended_remove()
        return cm

    return processor


@cli.command("triangulate")
@click.option(
    "--sloppy",
    is_flag=True,
    help="Use a more lenient triangulator (mapbox-earcut), which is also less robust.",
)
def triangulate_cmd(sloppy):
    """
    Triangulate every surface.

    If the robust method fails (crash) then it is caused by invalid input.
    You can use the option '--sloppy' which uses a more lenient library (mapbox-earcut),
    but watch out it is less robust (collapsed triangles could be created!).

    Takes care of updating: (1) semantics; (2) textures; (3) material.

    sage examples:

    \b
        cjio myfile.city.json triangulate save mytriangles.city.json
        cjio myfile.city.json triangulate --sloppy save mytriangles.city.json
    """

    # -- mapbox_earcut available?
    def processor(cm):
        print_cmd_status("Triangulate the CityJSON file")
        if not MODULE_TRIANGLE_AVAILABLE:
            str = "Cannot triangulate: Python module 'triangle' missing. Stopping here."
            print_cmd_alert(str)
            str = "Install it: https://pypi.org/project/triangle/"
            print_cmd_warning(str)
            raise click.ClickException("Abort.")
            return cm
        if not (cm.is_triangulated()):
            if sloppy and not MODULE_EARCUT_AVAILABLE:
                str = "Cannot triangulate: Python module 'mapbox_earcut' missing. Stopping here."
                print_cmd_alert(str)
                str = "Install it: https://pypi.org/project/mapbox-earcut/"
                print_cmd_warning(str)
                raise click.ClickException("Abort.")
            else:
                cm.triangulate(sloppy)
        else:
            print_cmd_status("This file is already triangulated!")
        return cm

    return processor


def _print_cmd(s, **styles):
    if isinstance(s, str):
        click.secho(s, **styles)
    elif isinstance(s, list):
        for w in s:
            if isinstance(w, warnings.WarningMessage):
                click.secho(w.message, **styles)
            else:
                raise TypeError(
                    "Can only print CLI warning from a string or a list of warning.WarningMessage"
                )
    else:
        raise TypeError(
            "Can only print CLI warning from a string or a list of warning.WarningMessage"
        )


@click.pass_context
def print_cmd_info(ctx, s):
    if not ctx.obj["suppress_msg"]:
        _print_cmd(s)


@click.pass_context
def print_cmd_status(ctx, s):
    if not ctx.obj["suppress_msg"]:
        _print_cmd(s, bg="cyan", fg="black")


@click.pass_context
def print_cmd_substatus(ctx, s):
    if not ctx.obj["suppress_msg"]:
        _print_cmd(s, fg="cyan")


@click.pass_context
def print_cmd_warning(ctx, s):
    if not ctx.obj["suppress_msg"]:
        _print_cmd(s, reverse=True, fg="yellow")


@click.pass_context
def print_cmd_alert(ctx, s):
    if not ctx.obj["suppress_msg"]:
        _print_cmd(s, reverse=True, fg="red")


# Needed for the executable created by PyInstaller
if getattr(sys, "frozen", False):
    cli(sys.argv[1:])



import os.path
import pytest
from click import ClickException

from cjio import utils


@pytest.fixture(
    scope="session",
    params=[
        ("doesnt_exist.json", False),
        ("tests/data/rotterdam", True),
        ("tests/data/rotterdam/", True),
        ("tests/data/delft.json", False),
        ("tests/data/doesnt_exist", True),
    ],
)
def valid_path(request):
    package_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    yield (os.path.join(package_dir, request.param[0]), request.param[1])


@pytest.fixture(
    scope="session",
    params=[
        ("tests/data/doesnt_exist/doesnt_exist.json", False),
        ("tests/data/doesnt_exist/other_dir", True),
    ],
)
def invalid_path(request):
    package_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    yield (os.path.join(package_dir, request.param[0]), request.param[1])


class TestFileNamesPaths:
    def test_verify_filenames_valid(self, valid_path):
        res = utils.verify_filename(valid_path[0])
        assert res["dir"] == valid_path[1]

    def test_verify_filenames_invalid(self, invalid_path):
        with pytest.raises(ClickException):
            utils.verify_filename(invalid_path[0])



__version__ = "0.10.0"
import importlib


loader = importlib.util.find_spec("triangle")
MODULE_TRIANGLE_AVAILABLE = loader is not None

loader = importlib.util.find_spec("mapbox_earcut")
MODULE_EARCUT_AVAILABLE = loader is not None

loader = importlib.util.find_spec("pyproj")
MODULE_PYPROJ_AVAILABLE = loader is not None

loader = importlib.util.find_spec("pandas")
MODULE_PANDAS_AVAILABLE = loader is not None

loader = importlib.util.find_spec("cjvalpy")
MODULE_CJVAL_AVAILABLE = loader is not None



"""CityModel subset functions"""


def select_co_ids(j, IDs):
    IDs = list(IDs)
    re = set()
    for theid in IDs:
        if theid in j["CityObjects"] and "parents" not in j["CityObjects"][theid]:
            re.add(theid)

    # deal with CityObjectGroup
    for each in j["CityObjects"]:
        if each in IDs:
            if (
                j["CityObjects"][each]["type"] == "CityObjectGroup"
                and "members" in j["CityObjects"][each]
            ):
                for member in j["CityObjects"][each]["members"]:
                    re.add(member)

    # -- add children "recursively"
    children_list = []
    for id in re:
        if "children" in j["CityObjects"][id]:
            for child in j["CityObjects"][id]["children"]:
                children_list.append(child)
    while len(children_list) > 0:
        c = children_list.pop()
        re.add(c)
        if "children" in j["CityObjects"][c]:
            for child in j["CityObjects"][c]["children"]:
                children_list.append(child)
    return re


def process_geometry(j, j2):
    # -- update vertex indices
    oldnewids = {}
    newvertices = []
    for each in j2["CityObjects"]:
        if "geometry" in j2["CityObjects"][each]:
            for geom in j2["CityObjects"][each]["geometry"]:
                update_array_indices(
                    geom["boundaries"], oldnewids, j["vertices"], newvertices, -1
                )
    j2["vertices"] = newvertices


def process_templates(j, j2):
    dOldNewIDs = {}
    newones = []
    for each in j2["CityObjects"]:
        if "geometry" in j2["CityObjects"][each]:
            for geom in j2["CityObjects"][each]["geometry"]:
                if geom["type"] == "GeometryInstance":
                    t = geom["template"]
                    if t in dOldNewIDs:
                        geom["template"] = dOldNewIDs[t]
                    else:
                        geom["template"] = len(newones)
                        dOldNewIDs[t] = len(newones)
                        newones.append(j["geometry-templates"]["templates"][t])
    if len(newones) > 0:
        j2["geometry-templates"] = {}
        j2["geometry-templates"]["vertices-templates"] = j["geometry-templates"][
            "vertices-templates"
        ]
        j2["geometry-templates"]["templates"] = newones


def process_appearance(j, j2):
    # -- materials
    dOldNewIDs = {}
    newmats = []
    for each in j2["CityObjects"]:
        if "geometry" in j2["CityObjects"][each]:
            for geom in j2["CityObjects"][each]["geometry"]:
                if "material" in geom:
                    for each in geom["material"]:
                        if "value" in geom["material"][each]:
                            v = geom["material"][each]["value"]
                            if v in dOldNewIDs:
                                geom["material"][each]["value"] = dOldNewIDs[v]
                            else:
                                geom["material"][each]["value"] = len(newmats)
                                dOldNewIDs[v] = len(newmats)
                                newmats.append(j["appearance"]["materials"][v])
                        if "values" in geom["material"][each]:
                            update_array_indices(
                                geom["material"][each]["values"],
                                dOldNewIDs,
                                j["appearance"]["materials"],
                                newmats,
                                -1,
                            )
    if len(newmats) > 0:
        j2["appearance"]["materials"] = newmats

    # -- textures references (first int in the arrays)
    dOldNewIDs = {}
    newtextures = []
    for each in j2["CityObjects"]:
        if "geometry" in j2["CityObjects"][each]:
            for geom in j2["CityObjects"][each]["geometry"]:
                if "texture" in geom:
                    for each in geom["texture"]:
                        if "values" in geom["texture"][each]:
                            update_array_indices(
                                geom["texture"][each]["values"],
                                dOldNewIDs,
                                j["appearance"]["textures"],
                                newtextures,
                                0,
                            )
    if len(newtextures) > 0:
        j2["appearance"]["textures"] = newtextures
    # -- textures vertices references (1+ int in the arrays)
    dOldNewIDs = {}
    newtextures = []
    for each in j2["CityObjects"]:
        if "geometry" in j2["CityObjects"][each]:
            for geom in j2["CityObjects"][each]["geometry"]:
                if "texture" in geom:
                    for each in geom["texture"]:
                        if "values" in geom["texture"][each]:
                            update_array_indices(
                                geom["texture"][each]["values"],
                                dOldNewIDs,
                                j["appearance"]["vertices-texture"],
                                newtextures,
                                1,
                            )
    if len(newtextures) > 0:
        j2["appearance"]["vertices-texture"] = newtextures


def update_array_indices(a, dOldNewIDs, oldarray, newarray, slicearray):
    # -- slicearray: -1=none ; 0=use-only-first (for textures) ; 1=use-1+ (for textures)
    # -- a must be an array
    # -- issue with passing integer is that it's non-mutable, thus can't update
    # -- (or I don't know how...)
    for i, each in enumerate(a):
        if isinstance(each, list):
            update_array_indices(each, dOldNewIDs, oldarray, newarray, slicearray)
        elif each is not None:
            if (
                (slicearray == -1)
                or (slicearray == 0 and i == 0)
                or (slicearray == 1 and i > 0)
            ):
                if each in dOldNewIDs:
                    a[i] = dOldNewIDs[each]
                else:
                    a[i] = len(newarray)
                    dOldNewIDs[each] = len(newarray)
                    newarray.append(oldarray[each])



"""Various utility functions"""

import os.path
import click


def verify_filename(filename):
    """Verify if the provided output filename is a file or a directory"""
    res = {"dir": False, "path": ""}
    if os.path.isdir(filename):
        res["dir"] = True
        absp = os.path.abspath(filename)
        if os.path.exists(absp):
            res["path"] = absp
        else:
            raise click.ClickException("Couldn't expand %s to absolute path" % filename)
    else:
        base = os.path.basename(filename)
        dirname = os.path.abspath(os.path.dirname(filename))
        # parent directory must exist, we don't recurse further
        if not os.path.exists(dirname):
            raise click.ClickException('Path does not exist: "%s"' % (dirname))
        fname, extension = os.path.splitext(base)
        res["path"] = os.path.join(dirname, base)
        if len(extension) == 0:
            res["dir"] = True
        else:
            res["dir"] = False
    return res



import math

import numpy as np

MODULE_TRIANGLE_AVAILABLE = True
try:
    import triangle
except ImportError:
    MODULE_TRIANGLE_AVAILABLE = False

MODULE_EARCUT_AVAILABLE = True
try:
    import mapbox_earcut
except ImportError:
    MODULE_EARCUT_AVAILABLE = False


def to_2d(p, n):
    # -- n must be normalised
    # p = np.array([1, 2, 3])
    # newell = np.array([1, 3, 4.2])
    # n = newell/math.sqrt(p[0]*p[0] + p[1]*p[1] + p[2]*p[2])
    x3 = np.array([1.1, 1.1, 1.1])  # -- is this always a good value??
    if (n == x3).all():
        x3 += np.array([1, 2, 3])
    x3 = x3 - np.dot(x3, n) * n
    # print(n, x3)
    x3 /= math.sqrt((x3**2).sum())  # make x a unit vector
    y3 = np.cross(n, x3)
    return (np.dot(p, x3), np.dot(p, y3))


def get_normal_newell(poly):
    """
    Compute the normal vector of a polygon using Newell's method.
    """
    n = np.array([0.0, 0.0, 0.0], dtype=np.float64)

    for i, p in enumerate(poly):
        if len(p) < 3:  # -- if it does not have 3 points then skip
            continue
        ne = i + 1
        if ne == len(poly):
            ne = 0
        n[0] += (poly[i][1] - poly[ne][1]) * (poly[i][2] + poly[ne][2])
        n[1] += (poly[i][2] - poly[ne][2]) * (poly[i][0] + poly[ne][0])
        n[2] += (poly[i][0] - poly[ne][0]) * (poly[i][1] + poly[ne][1])

    if (n == np.array([0.0, 0.0, 0.0])).all():
        # print("one wrong")
        return (n, False)
    n = n / math.sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2])
    return (n, True)


def triangulate_face(face, vnp, sloppy=False):
    if not sloppy:
        return triangulate_face_shewchuk(face, vnp)
    else:
        return triangulate_face_mapbox_earcut(face, vnp)


##-- with Shewchuk Triangle library
def triangulate_face_shewchuk(face, vnp):
    # print(face)
    # -- remove duplicate vertices, which can *easily* make Triangle segfault
    for i, each in enumerate(face):
        if len(set(each)) < len(each):  # -- there are duplicates
            re = []
            for k in each:
                if re.count(k) == 0:
                    re.append(k)
            face[i] = re
    # print(face)
    # print(len(face))
    if (len(face) == 1) and (len(face[0]) <= 3):
        if len(face[0]) == 3:
            # -- if a triangle then do nothing
            return (np.array(face), True)
        else:
            # -- if collapsed then ignore and return false
            return (np.array(face), False)

    for i, ring in enumerate(face):
        if len(ring) < 3:
            # -- if a triangle then do nothing
            return (np.zeros(1), False)
    sf = np.array([], dtype=np.int64)
    for ring in face:
        sf = np.hstack((sf, np.array(ring)))
    sfv = vnp[sf]

    rings = np.zeros(len(face), dtype=np.int64)
    total = 0
    for i in range(len(face)):
        total += len(face[i])
        rings[i] = total

    # 1. normal with Newell's method
    n, b = get_normal_newell(sfv)

    # 2. project to the plane to get xy
    sfv2d = np.zeros((sfv.shape[0], 2))
    for i, p in enumerate(sfv):
        xy = to_2d(p, n)
        sfv2d[i][0] = xy[0]
        sfv2d[i][1] = xy[1]

    # -- 3. deal with segments/constraints, prepare the Triangle input
    sg = np.zeros((rings[-1], 2), dtype=np.int64)
    for i, e in enumerate(sg):
        sg[i][0] = i
        sg[i][1] = i + 1
    starti = 0
    for each in rings:
        sg[each - 1][1] = starti
        starti = each
    # -- deal with holes
    if len(rings) > 1:
        holes = np.zeros((len(rings) - 1, 2))
        for k in range(len(rings) - 1):
            # -- basically triangulate the Triangle the Ring, and find the centre
            # -- of mass of the first triangle
            a = sfv2d[rings[k] : rings[k + 1]]
            sg1 = np.zeros((a.shape[0], 2), dtype=np.int64)
            for i, e in enumerate(sg1):
                sg1[i][0] = i
                sg1[i][1] = i + 1
            sg1[-1][1] = 0
            pcl = dict(vertices=a, segments=sg1)
            trl = triangle.triangulate(pcl, "p")
            t = trl["triangles"][0]
            c = np.average(a[t], axis=0)  # -- find its centre of mass
            holes[k][0] = c[0]
            holes[k][1] = c[1]
        A = dict(vertices=sfv2d, segments=sg, holes=holes)
    else:
        A = dict(vertices=sfv2d, segments=sg)

    try:
        re = triangle.triangulate(A, "p")
    except Exception as e:
        print(e)
        print("Houston we have a problem...")
        # re = {}
        return (np.array(face), False)
    # -- check the output
    if "triangles" not in re:
        return ([], False)
    re = re["triangles"]
    for i, each in enumerate(re):
        try:
            re[i] = sf[each]
        except Exception as exp:
            print(exp)
            return (re, False)
    return (re, True)


def triangulate_face_mapbox_earcut(face, vnp):
    sf = np.array([], dtype=np.int64)
    if (len(face) == 1) and (len(face[0]) == 3):
        return (np.array(face), True)
    for ring in face:
        sf = np.hstack((sf, np.array(ring)))
    sfv = vnp[sf]
    # print(sf)
    # print(sfv)
    rings = np.zeros(len(face), dtype=np.int32)
    total = 0
    for i in range(len(face)):
        total += len(face[i])
        rings[i] = total
    # print(rings)

    # 1. normal with Newell's method
    n, b = get_normal_newell(sfv)

    # -- if already a triangle then return it
    if not b:
        return (n, False)
    # print ("Newell:", n)

    # 2. project to the plane to get xy
    sfv2d = np.zeros((sfv.shape[0], 2))
    # print (sfv2d)
    for i, p in enumerate(sfv):
        xy = to_2d(p, n)
        # print("xy", xy)
        sfv2d[i][0] = xy[0]
        sfv2d[i][1] = xy[1]
    result = mapbox_earcut.triangulate_float64(sfv2d, rings)
    # print (result.reshape(-1, 3))

    for i, each in enumerate(result):
        # print (sf[i])
        result[i] = sf[each]

    # print (result.reshape(-1, 3))
    return (result.reshape(-1, 3), True)


def triangle_normal(tri, vertexlist, weighted=False):
    """Compute the triangle normal vector weighted by the triangle area.

    Returns None if the normal vector cannot be computed (mostly because the triangle
    is degenerate).
    """
    v0, v1, v2 = tri[0], tri[1], tri[2]
    p0 = np.array((vertexlist[v0][0], vertexlist[v0][1], vertexlist[v0][2]))
    p1 = np.array((vertexlist[v1][0], vertexlist[v1][1], vertexlist[v1][2]))
    p2 = np.array((vertexlist[v2][0], vertexlist[v2][1], vertexlist[v2][2]))
    cross_prod = np.cross(p1 - p0, p2 - p0)
    magnitude = np.linalg.norm(cross_prod)
    if math.isclose(magnitude, 0.0):
        return None
    else:
        norm_vec = cross_prod / magnitude
        if not weighted:
            return norm_vec
        else:
            tri_area = magnitude * 0.5
            return norm_vec * tri_area


def average_normal(normals):
    """Compute the smooth (average) normal vector from a list of vectors.
    Returns a numpy array of [x,y,z]. If the normal cannot be computed, then it returns
    a fake normal of [1.0, 0.0, 0.0].
    """
    s = sum(normals)
    n = np.linalg.norm(s)
    if math.isclose(n, 0.0):
        # Set a fake normal if length of the sum of normals is
        # 0. Can happen with opposite vectors.
        normal_vec = np.array([1.0, 0.0, 0.0])
    else:
        normal_vec = s / n
    return normal_vec



import copy
import json
import math
import os
import random
import re
import shutil
import sys
import urllib.request
import uuid
from datetime import datetime
from io import StringIO
from pathlib import Path

import numpy as np
from click import progressbar

from cjio import errors, convert, geom_help, subset
from cjio.errors import CJInvalidOperation
from cjio.floatEncoder import FloatEncoder

from cjio import (
    MODULE_PYPROJ_AVAILABLE,
    MODULE_CJVAL_AVAILABLE,
)

json.encoder.c_make_encoder = None
json.encoder.float = FloatEncoder

if MODULE_PYPROJ_AVAILABLE:
    from pyproj import CRS
    from pyproj.transformer import TransformerGroup

if MODULE_CJVAL_AVAILABLE:
    import cjvalpy


CITYJSON_VERSIONS_SUPPORTED = ["0.6", "0.8", "0.9", "1.0", "1.1", "2.0"]

CITYJSON_PROPERTIES = [
    "type",
    "version",
    "extensions",
    "transform",
    "metadata",
    "CityObjects",
    "vertices",
    "appearance",
    "geometry-templates",
]


def read_stdin():
    lcount = 1
    # -- read first line
    j1 = json.loads(sys.stdin.readline())
    cm = CityJSON(j=j1)
    if "CityObjects" not in cm.j:
        cm.j["CityObjects"] = {}
    if "vertices" not in cm.j:
        cm.j["vertices"] = []
    while True:
        lcount += 1
        line = sys.stdin.readline()
        if line != "":
            j1 = json.loads(line)
            if not ("type" in j1 and j1["type"] == "CityJSONFeature"):
                raise IOError(
                    "Line {} is not of type 'CityJSONFeature'.".format(lcount)
                )
            cm.add_cityjsonfeature(j1)
        else:
            cm.remove_duplicate_vertices()
            cm.remove_orphan_vertices()
            cm.update_bbox()
            break
    return cm


def reader(file, ignore_duplicate_keys=False):
    return CityJSON(file=file, ignore_duplicate_keys=ignore_duplicate_keys)


def off2cj(file):
    line = file.readline()
    while (len(line) <= 1) or (line[0] == "#") or (line[:3] == "OFF"):
        line = file.readline()
    numVertices = int(line.split()[0])
    numFaces = int(line.split()[1])
    lstVertices = []
    for i in range(numVertices):
        lstVertices.append(list(map(float, file.readline().split())))
    lstFaces = []
    for i in range(numFaces):
        lstFaces.append(list(map(int, file.readline().split()[1:])))
    cm = {}
    cm["type"] = "CityJSON"
    cm["version"] = CITYJSON_VERSIONS_SUPPORTED[-1]
    cm["CityObjects"] = {}
    cm["vertices"] = []
    for v in lstVertices:
        cm["vertices"].append(v)
    g = {"type": "Solid"}
    shell = []
    for f in lstFaces:
        shell.append([f])
    g["boundaries"] = [shell]
    g["lod"] = "1"
    o = {"type": "GenericCityObject"}
    o["geometry"] = [g]
    cm["CityObjects"]["id-1"] = o
    j = CityJSON(j=cm)
    j.compress()
    return j


def poly2cj(file):
    def read_next_line():
        """Helper function to skip comment lines and blank lines."""
        while True:
            line = file.readline()
            # Skip lines that are comments or blank
            if line and not line.strip().startswith("#") and line.strip():
                return line.split("#")[0].strip()
            if not line:  # End of file
                return None

    first_line = read_next_line()
    if first_line is None:
        raise ValueError("Invalid file: No content found.")
    numVertices = int(first_line.split()[0])

    lstVertices = []
    index = 0  # change depending the index used in the poly file
    for i in range(numVertices):
        line = read_next_line()
        if line is None:
            raise ValueError(
                "Invalid file: Unexpected end of file. while reading vertices."
            )

        if i == 0 and line.startswith("1"):
            index = -1
        lstVertices.append(list(map(float, line.split()))[1:])

    line = read_next_line()
    if line is None:
        raise ValueError(
            "Invalid file: Unexpected end of file while reading number of faces."
        )
    numFaces = int(line.split()[0])

    lstFaces = []
    for _ in range(numFaces):
        line = read_next_line()
        if line is None:
            raise ValueError(
                "Invalid file: Unexpected end of file while reading faces."
            )
        irings = int(line.split()[0]) - 1
        face = []
        face.append([x + index for x in map(int, read_next_line().split()[1:])])
        for _ in range(irings):
            face.append([x + index for x in map(int, read_next_line().split()[1:])])
        lstFaces.append(face)
    cm = {}
    cm["type"] = "CityJSON"
    cm["version"] = CITYJSON_VERSIONS_SUPPORTED[-1]
    cm["CityObjects"] = {}
    cm["vertices"] = []
    for v in lstVertices:
        cm["vertices"].append(v)
    g = {"type": "Solid"}
    shell = []
    for f in lstFaces:
        shell.append(f)
    g["boundaries"] = [shell]
    g["lod"] = "1"
    o = {"type": "GenericCityObject"}
    o["geometry"] = [g]
    cm["CityObjects"]["id-1"] = o
    j = CityJSON(j=cm)
    j.compress()
    return j


def update_geom_indices(a, offset):
    for i, each in enumerate(a):
        if isinstance(each, list):
            update_geom_indices(each, offset)
        else:
            if each is not None:
                a[i] = each + offset


def update_texture_indices(a, toffset, voffset):
    for i, each in enumerate(a):
        if isinstance(each, list):
            update_texture_indices(each, toffset, voffset)
        else:
            if each is not None:
                if i == 0:
                    a[i] = each + toffset
                else:
                    a[i] = each + voffset


class CityJSON:
    def __init__(self, file=None, j=None, ignore_duplicate_keys=False):
        if file is not None:
            self.read(file, ignore_duplicate_keys)
            self.path = os.path.abspath(file.name)
            self.reference_date = datetime.fromtimestamp(
                os.path.getmtime(file.name)
            ).strftime("%Y-%m-%d")
            self.cityobjects = {}
        elif j is not None:
            self.j = j
            self.cityobjects = {}
            self.path = None
            self.reference_date = datetime.now().strftime("%Y-%m-%d")
        else:  # -- create an empty one
            self.j = {}
            self.j["type"] = "CityJSON"
            self.j["version"] = CITYJSON_VERSIONS_SUPPORTED[-1]
            self.j["CityObjects"] = {}
            self.j["vertices"] = []
            self.cityobjects = {}
            self.path = None
            self.reference_date = datetime.now().strftime("%Y-%m-%d")

    def __repr__(self):
        return os.linesep.join(self.get_info())

    def get_version(self):
        return self.j["version"]

    def check_version(self):
        if not isinstance(self.get_version(), str):
            str1 = "CityJSON version should be a string 'X.Y' (eg '1.0')"
            raise errors.CJInvalidVersion(str1)
        pattern = re.compile(r"^(\d\.)(\d)$")  # -- correct pattern for version
        pattern2 = re.compile(r"^(\d\.)(\d\.)(\d)$")  # -- wrong pattern with X.Y.Z
        if pattern.fullmatch(self.get_version()) is None:
            if pattern2.fullmatch(self.get_version()) is not None:
                str1 = "CityJSON version should be only X.Y (eg '1.0') and not X.Y.Z (eg '1.0.1')"
                raise errors.CJInvalidVersion(str1)
            else:
                str1 = "CityJSON version is wrongly formatted"
                raise errors.CJInvalidVersion(str1)
        if self.get_version() not in CITYJSON_VERSIONS_SUPPORTED:
            allv = ""
            for v in CITYJSON_VERSIONS_SUPPORTED:
                allv = allv + v + "/"
            str1 = (
                "CityJSON version %s not supported (only versions: %s), not every operators will work.\nPerhaps it's time to upgrade cjio? 'pip install cjio -U'"
                % (self.get_version(), allv)
            )
            raise errors.CJInvalidVersion(str1)
        elif self.get_version() != CITYJSON_VERSIONS_SUPPORTED[-1]:
            str1 = (
                "v%s is not the latest version, and not everything will work.\n"
                % self.get_version()
            )
            str1 += "Upgrade the file with 'upgrade' command: 'cjio input.json upgrade save out.json'"
            errors.CJWarning(str1).warn()

    def get_epsg(self):
        if "metadata" not in self.j:
            return None
        if "referenceSystem" in self.j["metadata"]:
            s = self.j["metadata"]["referenceSystem"]
            if "opengis.net/def/crs" not in s or s.rfind("/") < 0:
                raise ValueError(
                    f"Invalid CRS string '{s}'. CRS needs to be formatted according to the OGC Name Type Specification: 'http://www.opengis.net/def/crs/{{authority}}/{{version}}/{{code}}'"
                )
            return int(s[s.rfind("/") + 1 :])
        else:
            return None

    def is_empty(self):
        if len(self.j["CityObjects"]) == 0:
            return True
        else:
            return False

    def is_transform(self):
        return "transform" in self.j

    def read(self, file, ignore_duplicate_keys=False):
        if ignore_duplicate_keys:
            try:
                self.j = json.loads(file.read())
            except json.decoder.JSONDecodeError as err:
                raise err
        else:
            try:
                self.j = json.loads(
                    file.read(), object_pairs_hook=self.dict_raise_on_duplicates
                )
            except ValueError as err:
                raise ValueError(err)
        # -- a CityJSON file?
        if "type" in self.j and self.j["type"] == "CityJSON":
            pass
        else:
            self.j = {}
            raise ValueError("Not a CityJSON file")

    def dict_raise_on_duplicates(self, ordered_pairs):
        d = {}
        for k, v in ordered_pairs:
            if k in d:
                raise ValueError(
                    "Invalid CityJSON file, duplicate key for City Object IDs: %r" % (k)
                )
            else:
                d[k] = v
        return d

    def validate(self):
        # -- only latest version, otherwise a mess with versions and different schemas
        # -- this is it, sorry people
        if self.j["version"] != CITYJSON_VERSIONS_SUPPORTED[-1]:
            s = (
                "Only files with version v%s can be validated. "
                % (CITYJSON_VERSIONS_SUPPORTED[-1])
            )
            raise Exception(s)
        val = cjvalpy.CJValidator(json.dumps(self.j))
        # -- fetch extensions from the URLs given
        js = []
        if "extensions" in self.j:
            for ext in self.j["extensions"]:
                # theurl = self.j["extensions"][ext]["url"]
                try:
                    with urllib.request.urlopen(self.j["extensions"][ext]["url"]) as f:
                        sf = f.read().decode("utf-8")
                        js.append(sf)
                except Exception as exp:
                    s = (
                        "'%s' cannot be downloaded\nAbort"
                        % self.j["extensions"][ext]["url"]
                    )
                    raise exp
        if not MODULE_CJVAL_AVAILABLE:
            raise ModuleNotFoundError(
                "Module 'cjvalpy' is not available, please install it"
            )
        val = cjvalpy.CJValidator(js)
        val.validate()
        re = val.get_report()
        return re

    def get_bbox(self):
        if "metadata" not in self.j:
            return self.calculate_bbox()
        if "geographicalExtent" not in self.j["metadata"]:
            return self.calculate_bbox()
        return self.j["metadata"]["geographicalExtent"]

    def calculate_bbox(self):
        """
        Calculate the bbox of the CityJSON.
        """
        if len(self.j["vertices"]) == 0:
            return [0, 0, 0, 0, 0, 0]
        x, y, z = zip(*self.j["vertices"])
        bbox = [min(x), min(y), min(z), max(x), max(y), max(z)]
        if "transform" in self.j:
            s = self.j["transform"]["scale"]
            t = self.j["transform"]["translate"]
            bbox = [a * b + c for a, b, c in zip(bbox, (s + s), (t + t))]
        return bbox

    def update_metadata(self):
        self.update_bbox()
        self.update_metadata_identifier()

    def update_bbox(self):
        """
        Update the bbox (["metadata"]["geographicalExtent"]) of the CityJSON.
        If there is none then it is added.
        """
        if "metadata" not in self.j:
            self.j["metadata"] = {}
        if self.is_empty():
            bbox = [0, 0, 0, 0, 0, 0]
            self.j["metadata"]["geographicalExtent"] = bbox
            return bbox
        bbox = self.calculate_bbox()
        self.j["metadata"]["geographicalExtent"] = bbox
        return bbox

    def update_metadata_identifier(self):
        """
        Replaces the metadata/identifier (if present) by a new one (UUID)
        """
        if "metadata" in self.j:
            if "identifier" in self.j["metadata"]:
                self.j["metadata"]["identifier"] = str(uuid.uuid4())

    def set_epsg(self, newepsg):
        if newepsg is None:
            if "metadata" in self.j:
                if "referenceSystem" in self.j["metadata"]:
                    del self.j["metadata"]["referenceSystem"]
            return True
        try:
            i = int(newepsg)
        except ValueError:
            return False
        if "metadata" not in self.j:
            self.j["metadata"] = {}
        if "referenceSystem" not in self.j["metadata"]:
            self.j["metadata"]["referenceSystem"] = {}
        s = "https://www.opengis.net/def/crs/EPSG/0/" + str(i)
        self.j["metadata"]["referenceSystem"] = s
        return True

    def update_bbox_each_cityobjects(self, addifmissing=False):
        def recusionvisit(a, vs):
            for each in a:
                if isinstance(each, list):
                    recusionvisit(each, vs)
                else:
                    vs.append(each)

        for co in self.j["CityObjects"]:
            if addifmissing or "geographicalExtent" in self.j["CityObjects"][co]:
                vs = []
                bbox = [9e9, 9e9, 9e9, -9e9, -9e9, -9e9]
                if "geometry" in self.j["CityObjects"][co]:
                    for g in self.j["CityObjects"][co]["geometry"]:
                        recusionvisit(g["boundaries"], vs)
                        for each in vs:
                            v = self.j["vertices"][each]
                            for i in range(3):
                                if v[i] < bbox[i]:
                                    bbox[i] = v[i]
                            for i in range(3):
                                if v[i] > bbox[i + 3]:
                                    bbox[i + 3] = v[i]
                        if "transform" in self.j:
                            for i in range(3):
                                bbox[i] = (
                                    bbox[i] * self.j["transform"]["scale"][i]
                                ) + self.j["transform"]["translate"][i]
                            for i in range(3):
                                bbox[i + 3] = (
                                    bbox[i + 3] * self.j["transform"]["scale"][i]
                                ) + self.j["transform"]["translate"][i]
                        self.j["CityObjects"][co]["geographicalExtent"] = bbox

    def get_centroid(self, coid):
        def recusionvisit(a, vs):
            for each in a:
                if isinstance(each, list):
                    recusionvisit(each, vs)
                else:
                    vs.append(each)

        # -- find the 3D centroid
        centroid = [0, 0, 0]
        total = 0
        if "geometry" in self.j["CityObjects"][coid]:
            for g in self.j["CityObjects"][coid]["geometry"]:
                vs = []
                recusionvisit(g["boundaries"], vs)
                for each in vs:
                    v = self.j["vertices"][each]
                    total += 1
                    centroid[0] += v[0]
                    centroid[1] += v[1]
                    centroid[2] += v[2]
        if total != 0:
            centroid[0] /= total
            centroid[1] /= total
            centroid[2] /= total
            if "transform" in self.j:
                centroid[0] = (centroid[0] * self.j["transform"]["scale"][0]) + self.j[
                    "transform"
                ]["translate"][0]
                centroid[1] = (centroid[1] * self.j["transform"]["scale"][1]) + self.j[
                    "transform"
                ]["translate"][1]
                centroid[2] = (centroid[2] * self.j["transform"]["scale"][2]) + self.j[
                    "transform"
                ]["translate"][2]
            return centroid
        else:
            return None

    def get_identifier(self):
        """
        Returns the identifier of this file.

        If there is one in metadata, it will be returned. Otherwise, the filename will.
        """
        if "metadata" in self.j:
            if "identifier" in self.j["metadata"]:
                cm_id = self.j["metadata"]["identifier"]
        if cm_id:
            template = "{cm_id} ({file_id})"
        else:
            template = "{file_id}"
        if "metadata" in self.j:
            if "identifier" in self.j["metadata"]:
                return template.format(
                    cm_id=cm_id, file_id=self.j["metadata"]["identifier"]
                )
        if self.path:
            return os.path.basename(self.path)
        return "unknown"

    def get_title(self):
        """
        Returns the description of this file from metadata.

        If there is none, the identifier will be returned, instead.
        """

        if "metadata" in self.j:
            if "title" in self.j["metadata"]:
                return self.j["metadata"]["title"]

        return self.get_identifier()

    def get_subset_bbox(self, bbox, exclude=False):
        re = set()
        for coid in self.j["CityObjects"]:
            centroid = self.get_centroid(coid)
            if (
                (centroid is not None)
                and (centroid[0] >= bbox[0])
                and (centroid[1] >= bbox[1])
                and (centroid[0] < bbox[2])
                and (centroid[1] < bbox[3])
            ):
                re.add(coid)
        return self.subset(lsIDs=re, exclude=exclude)

    def get_subset_radius(self, x, y, radius, exclude=False):
        re = set()
        for coid in self.j["CityObjects"]:
            centroid = self.get_centroid(coid)
            if centroid is not None:
                dist = ((centroid[0] - x) ** 2) + ((centroid[1] - y) ** 2)
                if dist < radius**2:
                    re.add(coid)
        return self.subset(lsIDs=re, exclude=exclude)

    def is_co_toplevel(self, co):
        return "parents" not in co

    def subset(self, lsIDs, exclude=False):
        # -- copy selected CO to the j2
        re = subset.select_co_ids(self.j, lsIDs)
        # -- exclude
        if exclude:
            sallkeys = set(self.j["CityObjects"].keys())
            re = sallkeys - re
        re = list(re)
        # -- new sliced CityJSON object
        cm2 = CityJSON()
        cm2.j["version"] = copy.deepcopy(self.j["version"])
        cm2.path = copy.deepcopy(self.path)
        if "extensions" in self.j:
            cm2.j["extensions"] = copy.deepcopy(self.j["extensions"])
        if "transform" in self.j:
            cm2.j["transform"] = copy.deepcopy(self.j["transform"])
        # -- select only the COs
        for each in re:
            cm2.j["CityObjects"][each] = copy.deepcopy(self.j["CityObjects"][each])
        # -- geometry
        subset.process_geometry(self.j, cm2.j)
        # -- templates
        subset.process_templates(self.j, cm2.j)
        # -- appearance
        if "appearance" in self.j:
            cm2.j["appearance"] = {}
            subset.process_appearance(self.j, cm2.j)
        # -- copy all other non mandatory properties
        for p in self.j:
            if p not in CITYJSON_PROPERTIES:
                cm2.j[p] = copy.deepcopy(self.j[p])
        # -- metadata
        if "metadata" in self.j:
            cm2.j["metadata"] = copy.deepcopy(self.j["metadata"])
            cm2.update_metadata()
        return cm2

    def get_subset_random(self, number=1, exclude=False):
        """Get a random sample of CityObjects without replacement."""
        top_level_cos = [
            id for id, co in self.j["CityObjects"].items() if self.is_co_toplevel(co)
        ]
        try:
            random_ids = random.sample(top_level_cos, k=number)
        except ValueError:
            # If the sample size 'k' is larger than the number of CityObjects
            random_ids = top_level_cos
        return self.subset(lsIDs=random_ids, exclude=exclude)

    def get_subset_ids(self, lsIDs, exclude=False):
        return self.subset(lsIDs=set(lsIDs), exclude=exclude)

    def get_subset_cotype(self, cotypes, exclude=False):
        # random.seed()
        # total = len(self.j["CityObjects"])
        # if number > total:
        #     number = total
        # allkeys = list(self.j["CityObjects"].keys())
        # re = set()
        # count = 0
        # while (count < number):
        #     t = allkeys[random.randint(0, total - 1)]
        #     if self.is_co_toplevel(self.j["CityObjects"][t]):
        #         re.add(t)
        #         count += 1
        # return self.subset(lsIDs=re, exclude=exclude)
        re = set()
        for theid in self.j["CityObjects"]:
            if self.j["CityObjects"][theid]["type"] in cotypes:
                re.add(theid)
        return self.subset(lsIDs=re, exclude=exclude)

    def get_textures_location(self):
        """Get the location of the texture files

        Assumes that all textures are in the same location. Relative paths
        are expanded to absolute paths.

        :returns: path to the directory or URL of the texture files
        :rtype: string (path) or None (on failure)
        :raises: NotADirectoryError
        """
        if "appearance" in self.j:
            if "textures" in self.j["appearance"]:
                p = self.j["appearance"]["textures"][0]["image"]
                cj_dir = os.path.dirname(self.path)
                url = re.match(r"http[s]?://|www\.", p)
                if url:
                    return url
                else:
                    texture_dir = os.path.dirname(p)
                    if len(texture_dir) == 0:
                        # textures are in the same dir as the cityjson file
                        return cj_dir
                    else:  # if path is empty
                        # if absolute and exists return it
                        if os.path.isabs(texture_dir) and os.path.isdir(texture_dir):
                            return texture_dir
                        # if relative and its absolute exists return it
                        elif not os.path.isabs(texture_dir) and os.path.isdir(
                            os.path.abspath(texture_dir)
                        ):
                            return os.path.abspath(texture_dir)
                        # if relative and concatenation with cj_dir exists return it
                        elif not os.path.isabs(texture_dir) and os.path.isdir(
                            os.path.join(cj_dir, texture_dir)
                        ):
                            return os.path.join(cj_dir, texture_dir)
                        else:
                            raise NotADirectoryError(
                                "Texture directory '%s' not found" % texture_dir
                            )
            else:
                return None
        else:
            return None

    def update_textures_location(self, new_loc, relative=True):
        """Updates the location of the texture files

        If the new location is a directory in the local file system, it is
        expected to exists with the texture files in it.

        :param new_loc: path to new texture directory
        :type new_loc: string
        :param relative: create texture links relative to the CityJSON file
        :type relative: boolean
        :returns: None -- modifies the CityJSON
        :raises: InvalidOperation, NotADirectoryError
        """
        curr_loc = self.get_textures_location()
        if curr_loc:
            if re.match(r"http[s]?://|www\.", new_loc):
                apath = new_loc
                for t in self.j["appearance"]["textures"]:
                    f = os.path.basename(t["image"])
                    t["image"] = os.path.join(apath, f)
            else:
                apath = os.path.abspath(new_loc)
                if not os.path.isdir(apath):
                    raise NotADirectoryError("%s does not exits" % apath)
                elif relative:
                    rpath = os.path.relpath(apath, os.path.dirname(self.path))
                    for t in self.j["appearance"]["textures"]:
                        f = os.path.basename(t["image"])
                        t["image"] = os.path.join(rpath, f)
                else:
                    for t in self.j["appearance"]["textures"]:
                        f = os.path.basename(t["image"])
                        t["image"] = os.path.join(apath, f)
        else:
            raise CJInvalidOperation(
                "Cannot update textures in a city model without textures"
            )

    def copy_textures(self, new_textures_loc: str) -> None:
        """Copy the texture files to a new location
        :param new_textures_loc: path to new texture directory
        :raises: InvalidOperation, IOError
        """
        current_textures_loc = self.get_textures_location()
        if not current_textures_loc:
            raise CJInvalidOperation(
                "Cannot copy textures from a city model without textures"
            )

        # get the absolute path of the new textures location
        new_textures_abs_loc = os.path.abspath(new_textures_loc)

        # create the new textures directory if it does not exist
        if not os.path.isdir(new_textures_abs_loc):
            os.mkdir(new_textures_abs_loc)

        try:
            for t in self.j["appearance"]["textures"]:
                f = os.path.basename(t["image"])
                current_texture_file_path = os.path.join(current_textures_loc, f)
                # if the file exist, copy it to the new location
                if os.path.isfile(current_texture_file_path):
                    shutil.copy(current_texture_file_path, new_textures_abs_loc)
            # update the location relative to the CityJSON file
            self.update_textures_location(new_textures_abs_loc, relative=False)
            print("Textures copied to", new_textures_abs_loc)
        except IOError as e:
            raise IOError(" Error copying texture files: %s" % e)

    def validate_textures(self):
        """Check if the texture files exist"""
        # TODO: implement validate_textures
        raise NotImplementedError

    def remove_textures(self):
        for co in self.j["CityObjects"].values():
            for geom in co.get("geometry", []):
                if "texture" in geom:
                    del geom["texture"]
        if "appearance" in self.j:
            if "textures" in self.j["appearance"]:
                del self.j["appearance"]["textures"]
            if "vertices-texture" in self.j["appearance"]:
                del self.j["appearance"]["vertices-texture"]
            if "default-theme-texture" in self.j["appearance"]:
                del self.j["appearance"]["default-theme-texture"]
        if "appearance" in self.j:
            if self.j["appearance"] is None or len(self.j["appearance"]) == 0:
                del self.j["appearance"]
        return True

    def remove_materials(self):
        for co in self.j["CityObjects"].values():
            for geom in co.get("geometry", []):
                if "material" in geom:
                    del geom["material"]
        if "appearance" in self.j:
            if "materials" in self.j["appearance"]:
                del self.j["appearance"]["materials"]
            if "default-theme-material" in self.j["appearance"]:
                del self.j["appearance"]["default-theme-material"]
        if "appearance" in self.j:
            if self.j["appearance"] is None or len(self.j["appearance"]) == 0:
                del self.j["appearance"]
        return True

    def number_city_objects_level1(self):
        total = 0
        for id in self.j["CityObjects"]:
            if self.is_co_toplevel(self.j["CityObjects"][id]):
                total += 1
        return total

    def get_info(self, long=False):
        s = []
        s.append("CityJSON version = {}".format(self.get_version()))
        s.append("EPSG = {}".format(self.get_epsg()))
        sb = ""
        for each in self.get_bbox():
            sb += "{:.3f} ".format(each)
        s.append("bbox = [ {}]".format(sb))
        if "extensions" in self.j:
            d = set()
            for i in self.j["extensions"]:
                d.add(i)
            s.append("Extensions = {}".format(sorted(list(d))))
        # -- hierarchy tree for CityObjects
        s.append("=== CityObjects ===")
        d = {}
        for key in self.j["CityObjects"]:
            ty = self.j["CityObjects"][key]["type"]
            if "parents" not in self.j["CityObjects"][key]:
                if ty not in d:
                    d[ty] = 1
                else:
                    d[ty] += 1
                self.info_children_dfs(key, ty, d)
        for each in d:
            if each.count("/") == 0:
                s2 = "|-- {} ({})".format(each, d[each])
                s.append(s2)
                self.print_info_tree(s, d, each, 1)
        s.append("===================")
        if "appearance" in self.j:
            s.append("materials = {}".format("materials" in self.j["appearance"]))
            s.append("textures = {}".format("textures" in self.j["appearance"]))
        else:
            s.append("materials = {}".format(False))
            s.append("textures = {}".format(False))
        if not long:
            return s
        # -- all/long version
        s.append("vertices_total = {}".format(len(self.j["vertices"])))
        s.append("is_triangulated = {}".format(self.is_triangulated()))
        s.append("transform/scale = {}".format(self.j["transform"]["scale"]))
        s.append("transform/translate = {}".format(self.j["transform"]["translate"]))
        geoms = set()
        lod = set()
        sem_srf = set()
        co_attributes = set()
        for key in self.j["CityObjects"]:
            if "attributes" in self.j["CityObjects"][key]:
                for attr in self.j["CityObjects"][key]["attributes"].keys():
                    co_attributes.add(attr)
            if "geometry" in self.j["CityObjects"][key]:
                for geom in self.j["CityObjects"][key]["geometry"]:
                    geoms.add(geom["type"])
                    if "lod" in geom:
                        lod.add(geom["lod"])
                    else:  # -- it's a geometry-template
                        lod.add(
                            self.j["geometry-templates"]["templates"][geom["template"]][
                                "lod"
                            ]
                        )
                    if "semantics" in geom:
                        if geom["semantics"] is not None:
                            for srf in geom["semantics"]["surfaces"]:
                                sem_srf.add(srf["type"])

        def getsorted(a):
            return sorted(list(a))

        s.append("geom primitives = {}".format(getsorted(geoms)))
        s.append("LoD = {}".format(getsorted(lod)))
        s.append("semantics surfaces = {}".format(getsorted(sem_srf)))
        s.append("attributes = {}".format(getsorted(co_attributes)))
        return s

    def print_info_tree(self, s, d, t, level):
        for each in d:
            if each.startswith(t + "/") and each.count("/") == level:
                x = each.rsplit("/")[-1]
                s2 = "{}|-- {} ({})".format(" " * 4 * level, x, d[each])
                s.append(s2)
                self.print_info_tree(s, d, each, level + 1)

    def info_children_dfs(self, key, typeparent, d):
        if "children" in self.j["CityObjects"][key]:
            for c in self.j["CityObjects"][key]["children"]:
                ct = self.j["CityObjects"][c]["type"]
                s = typeparent + "/" + ct
                if s not in d:
                    d[s] = 1
                else:
                    d[s] += 1
                self.info_children_dfs(c, s, d)

    def remove_orphan_vertices(self):
        def visit_geom(a, oldnewids, newvertices):
            for i, each in enumerate(a):
                if isinstance(each, list):
                    visit_geom(each, oldnewids, newvertices)
                else:
                    if each not in oldnewids:
                        oldnewids[each] = len(newvertices)
                        newvertices.append(each)

        def update_face(a, oldnewids):
            for i, each in enumerate(a):
                if isinstance(each, list):
                    update_face(each, oldnewids)
                else:
                    a[i] = oldnewids[each]

        # --
        totalinput = len(self.j["vertices"])
        oldnewids = {}
        newvertices = []
        # -- visit each geom to gather used ids
        for theid in self.j["CityObjects"]:
            if "geometry" in self.j["CityObjects"][theid]:
                for g in self.j["CityObjects"][theid]["geometry"]:
                    visit_geom(g["boundaries"], oldnewids, newvertices)
        # -- update the faces ids
        for theid in self.j["CityObjects"]:
            if "geometry" in self.j["CityObjects"][theid]:
                for g in self.j["CityObjects"][theid]["geometry"]:
                    update_face(g["boundaries"], oldnewids)
        # -- replace the vertices, innit?
        newv2 = []
        for v in newvertices:
            newv2.append(self.j["vertices"][v])
        self.j["vertices"] = newv2
        return totalinput - len(self.j["vertices"])

    def remove_duplicate_vertices(self):
        def update_geom_indices(a, newids):
            for i, each in enumerate(a):
                if isinstance(each, list):
                    update_geom_indices(each, newids)
                else:
                    a[i] = newids[each]

        # --
        totalinput = len(self.j["vertices"])
        h = {}
        newids = [-1] * len(self.j["vertices"])
        newvertices = []
        for i, v in enumerate(self.j["vertices"]):
            s = "{x} {y} {z}".format(x=v[0], y=v[1], z=v[2])
            if s not in h:
                newid = len(h)
                newids[i] = newid
                h[s] = newid
                newvertices.append(s)
            else:
                newids[i] = h[s]
        # -- update indices
        for theid in self.j["CityObjects"]:
            if "geometry" in self.j["CityObjects"][theid]:
                for g in self.j["CityObjects"][theid]["geometry"]:
                    update_geom_indices(g["boundaries"], newids)
        # -- replace the vertices, innit?
        newv2 = []
        for v in newvertices:
            if "transform" in self.j:
                a = list(map(int, v.split()))
            else:
                a = list(map(float, v.split()))
            newv2.append(a)
        self.j["vertices"] = newv2
        return totalinput - len(self.j["vertices"])

    def compress(self, important_digits=3, translate=None):
        """Compress the city model by scaling and translating it.

        The scaling factor is defined by 'important_digits'. The translation properties
        are either determined as the minimum coordinates of the city model if
        'translate=None', or can be
        set by providing the ``[x, y, z]`` translation properties to 'translate'.
        """
        if "transform" in self.j:
            return False
        # -- find the minx/miny/minz or set from translate
        if translate:
            bbox = translate
        else:
            bbox = [9e9, 9e9, 9e9]
            for v in self.j["vertices"]:
                for i in range(3):
                    if v[i] < bbox[i]:
                        bbox[i] = v[i]
        # convert vertices in self.j to int
        n = [0, 0, 0]
        p = "%." + str(important_digits) + "f"
        for v in self.j["vertices"]:
            for i in range(3):
                n[i] = v[i] - bbox[i]
            for i in range(3):
                v[i] = int((p % n[i]).replace(".", ""))

        # put transform
        ss = 1.0 / (math.pow(10, important_digits))

        self.j["transform"] = {}
        self.j["transform"]["scale"] = [ss, ss, ss]
        self.j["transform"]["translate"] = [bbox[0], bbox[1], bbox[2]]
        # -- clean the file
        _ = self.remove_duplicate_vertices()
        _ = self.remove_orphan_vertices()
        return True

    def decompress(self):
        if "transform" in self.j:
            for v in self.j["vertices"]:
                v[0] = (v[0] * self.j["transform"]["scale"][0]) + self.j["transform"][
                    "translate"
                ][0]
                v[1] = (v[1] * self.j["transform"]["scale"][1]) + self.j["transform"][
                    "translate"
                ][1]
                v[2] = (v[2] * self.j["transform"]["scale"][2]) + self.j["transform"][
                    "translate"
                ][2]
            del self.j["transform"]
            return True
        else:
            return False

    def add_cityjsonfeature(self, j):
        offset = len(self.j["vertices"])
        self.j["vertices"] += j["vertices"]
        # -- add each CityObjects
        for theid in j["CityObjects"]:
            self.j["CityObjects"][theid] = j["CityObjects"][theid]
            if "geometry" in self.j["CityObjects"][theid]:
                for g in self.j["CityObjects"][theid]["geometry"]:
                    update_geom_indices(g["boundaries"], offset)

        # -- materials
        if ("appearance" in j) and ("materials" in j["appearance"]):
            if ("appearance" in self.j) and ("materials" in self.j["appearance"]):
                offset = len(self.j["appearance"]["materials"])
            else:
                if "appearance" not in self.j:
                    self.j["appearance"] = {}
                if "materials" not in self.j["appearance"]:
                    self.j["appearance"]["materials"] = []
                offset = 0
            # -- copy materials
            for m in j["appearance"]["materials"]:
                self.j["appearance"]["materials"].append(m)
            # -- update the "material" in each Geometry
            for theid in j["CityObjects"]:
                for g in self.j["CityObjects"][theid].get("geometry", []):
                    if "material" in g:
                        for m in g["material"]:
                            if "values" in g["material"][m]:
                                update_geom_indices(g["material"][m]["values"], offset)
                            else:
                                g["material"][m]["value"] = (
                                    g["material"][m]["value"] + offset
                                )
        # -- textures
        if ("appearance" in j) and ("textures" in j["appearance"]):
            if ("appearance" in self.j) and ("textures" in self.j["appearance"]):
                toffset = len(self.j["appearance"]["textures"])
                voffset = len(self.j["appearance"]["vertices-texture"])
            else:
                if "appearance" not in self.j:
                    self.j["appearance"] = {}
                if "textures" not in self.j["appearance"]:
                    self.j["appearance"]["textures"] = []
                if "vertices-texture" not in self.j["appearance"]:
                    self.j["appearance"]["vertices-texture"] = []
                toffset = 0
                voffset = 0
            # -- copy vertices-texture
            self.j["appearance"]["vertices-texture"] += j["appearance"][
                "vertices-texture"
            ]
            # -- copy textures
            for t in j["appearance"]["textures"]:
                self.j["appearance"]["textures"].append(t)
            # -- update the "texture" in each Geometry
            for theid in j["CityObjects"]:
                for g in self.j["CityObjects"][theid]["geometry"]:
                    if "texture" in g:
                        for m in g["texture"]:
                            if "values" in g["texture"][m]:
                                update_texture_indices(
                                    g["texture"][m]["values"], toffset, voffset
                                )
                            else:
                                raise KeyError(
                                    f"The member 'values' is missing from the texture '{m}' in CityObject {theid}"
                                )

        # self.remove_duplicate_vertices()
        # self.remove_orphan_vertices()
        # self.update_bbox()

    def merge(self, lsCMs):
        # decompress() everything
        # updates CityObjects
        # updates vertices
        # updates geometry-templates
        # updates textures
        # updates materials
        #############################

        # -- decompress current CM
        imp_digits = math.ceil(abs(math.log(self.j["transform"]["scale"][0], 10)))
        self.decompress()

        for cm in lsCMs:
            # -- decompress
            if math.ceil(abs(math.log(cm.j["transform"]["scale"][0], 10))) > imp_digits:
                imp_digits = math.ceil(abs(math.log(cm.j["transform"]["scale"][0], 10)))
            cm.decompress()
            offset = len(self.j["vertices"])
            self.j["vertices"] += cm.j["vertices"]
            # -- add each CityObjects
            for theid in cm.j["CityObjects"]:
                if theid in self.j["CityObjects"]:
                    # -- merge attributes if not present (based on the property name only)
                    if "attributes" in cm.j["CityObjects"][theid]:
                        for a in cm.j["CityObjects"][theid]["attributes"]:
                            if a not in self.j["CityObjects"][theid]["attributes"]:
                                self.j["CityObjects"][theid]["attributes"][a] = cm.j[
                                    "CityObjects"
                                ][theid]["attributes"][a]
                    # -- merge geoms if not present (based on LoD only)
                    if "geometry" in self.j["CityObjects"][theid]:
                        for g in cm.j["CityObjects"][theid]["geometry"]:
                            thelod = str(g["lod"])
                            b = False
                            for g2 in self.j["CityObjects"][theid]["geometry"]:
                                if g2["lod"] == thelod:
                                    b = True
                                    break
                            if not b:
                                self.j["CityObjects"][theid]["geometry"].append(g)
                                update_geom_indices(
                                    self.j["CityObjects"][theid]["geometry"][-1][
                                        "boundaries"
                                    ],
                                    offset,
                                )
                else:
                    # -- copy the CO
                    self.j["CityObjects"][theid] = cm.j["CityObjects"][theid]
                    if "geometry" in self.j["CityObjects"][theid]:
                        for g in self.j["CityObjects"][theid]["geometry"]:
                            update_geom_indices(g["boundaries"], offset)
            # -- templates
            if "geometry-templates" in cm.j:
                if "geometry-templates" in self.j:
                    notemplates = len(self.j["geometry-templates"]["templates"])
                    novtemplate = len(
                        self.j["geometry-templates"]["vertices-templates"]
                    )
                else:
                    self.j["geometry-templates"] = {}
                    self.j["geometry-templates"]["templates"] = []
                    self.j["geometry-templates"]["vertices-templates"] = []
                    notemplates = 0
                    novtemplate = 0
                # -- copy templates
                for t in cm.j["geometry-templates"]["templates"]:
                    self.j["geometry-templates"]["templates"].append(t)
                    tmp = self.j["geometry-templates"]["templates"][-1]
                    update_geom_indices(tmp["boundaries"], novtemplate)
                # -- copy vertices
                self.j["geometry-templates"]["vertices-templates"] += cm.j[
                    "geometry-templates"
                ]["vertices-templates"]
                # -- update the "template" in each GeometryInstance
                for theid in cm.j["CityObjects"]:
                    if "geometry" in self.j["CityObjects"][theid]:
                        for g in self.j["CityObjects"][theid]["geometry"]:
                            if g["type"] == "GeometryInstance":
                                g["template"] += notemplates
            # -- materials
            if ("appearance" in cm.j) and ("materials" in cm.j["appearance"]):
                if ("appearance" in self.j) and ("materials" in self.j["appearance"]):
                    offset = len(self.j["appearance"]["materials"])
                else:
                    if "appearance" not in self.j:
                        self.j["appearance"] = {}
                    if "materials" not in self.j["appearance"]:
                        self.j["appearance"]["materials"] = []
                    offset = 0
                # -- copy materials
                for m in cm.j["appearance"]["materials"]:
                    self.j["appearance"]["materials"].append(m)
                # -- update the "material" in each Geometry
                for theid in cm.j["CityObjects"]:
                    for g in self.j["CityObjects"][theid].get("geometry", []):
                        if "material" in g:
                            for m in g["material"]:
                                if "values" in g["material"][m]:
                                    update_geom_indices(
                                        g["material"][m]["values"], offset
                                    )
                                else:
                                    g["material"][m]["value"] = (
                                        g["material"][m]["value"] + offset
                                    )
            # -- textures
            if ("appearance" in cm.j) and ("textures" in cm.j["appearance"]):
                if ("appearance" in self.j) and ("textures" in self.j["appearance"]):
                    toffset = len(self.j["appearance"]["textures"])
                    voffset = len(self.j["appearance"]["vertices-texture"])
                else:
                    if "appearance" not in self.j:
                        self.j["appearance"] = {}
                    if "textures" not in self.j["appearance"]:
                        self.j["appearance"]["textures"] = []
                    if "vertices-texture" not in self.j["appearance"]:
                        self.j["appearance"]["vertices-texture"] = []
                    toffset = 0
                    voffset = 0
                # -- copy vertices-texture
                self.j["appearance"]["vertices-texture"] += cm.j["appearance"][
                    "vertices-texture"
                ]
                # -- copy textures
                for t in cm.j["appearance"]["textures"]:
                    self.j["appearance"]["textures"].append(t)
                # -- update the "texture" in each Geometry
                for theid in cm.j["CityObjects"]:
                    for g in self.j["CityObjects"][theid].get("geometry", []):
                        if "texture" in g:
                            for m in g["texture"]:
                                if "values" in g["texture"][m]:
                                    update_texture_indices(
                                        g["texture"][m]["values"], toffset, voffset
                                    )
                                else:
                                    raise KeyError(
                                        f"The member 'values' is missing from the texture '{m}' in CityObject {theid}"
                                    )

        self.remove_duplicate_vertices()
        self.remove_orphan_vertices()
        self.update_bbox()
        self.compress(imp_digits)
        return True

    def upgrade_version_v06_v08(self):
        # -- version
        self.j["version"] = "0.8"
        # -- crs/epgs
        epsg = self.get_epsg()
        self.j["metadata"] = {}
        if epsg is not None:
            if "crs" in self.j["metadata"]:
                del self.j["metadata"]["crs"]
            self.set_epsg(epsg)
        # -- bbox
        self.update_bbox()
        for id in self.j["CityObjects"]:
            if "bbox" in self.j["CityObjects"][id]:
                self.j["CityObjects"][id]["geographicalExtent"] = self.j["CityObjects"][
                    id
                ]["bbox"]
                del self.j["CityObjects"][id]["bbox"]
        # #-- parent-children: do children have the parent too?
        subs = ["Parts", "Installations", "ConstructionElements"]
        for id in self.j["CityObjects"]:
            children = []
            for sub in subs:
                if sub in self.j["CityObjects"][id]:
                    for each in self.j["CityObjects"][id][sub]:
                        children.append(each)
            if len(children) > 0:
                # -- remove the Parts/Installations
                self.j["CityObjects"][id]["children"] = children
                for sub in subs:
                    if sub in self.j["CityObjects"][id]:
                        del self.j["CityObjects"][id][sub]
                # -- put the "parent" in each children
                for child in children:
                    if child in self.j["CityObjects"]:
                        self.j["CityObjects"][child]["parent"] = id

    def upgrade_version_v08_v09(self, reasons):
        # -- version
        self.j["version"] = "0.9"
        # -- parent --> parents[]
        allids = []
        for id in self.j["CityObjects"]:
            allids.append(id)
        for id in allids:
            if "parent" in self.j["CityObjects"][id]:
                self.j["CityObjects"][id]["parents"] = [
                    self.j["CityObjects"][id]["parent"]
                ]
                del self.j["CityObjects"][id]["parent"]
        # -- extensions
        if "extensions" in self.j:
            reasons += (
                "Extensions have changed completely in v0.9, update them manually."
            )
            return (False, reasons)
        return (True, "")

    def upgrade_version_v09_v10(self, reasons):
        # -- version
        self.j["version"] = "1.0"
        # -- extensions
        if "extensions" in self.j:
            for ext in self.j["extensions"]:
                theurl = self.j["extensions"][ext]
                self.j["extensions"][ext] = {"url": theurl, "version": "0.1"}
        return (True, "")

    def upgrade_version_v10_v11(self, reasons, digit):
        # -- version
        self.j["version"] = "1.1"
        # -- compress for "transform"
        self.compress(digit)
        # -- lod=string
        for theid in self.j["CityObjects"]:
            if "geometry" in self.j["CityObjects"][theid]:
                for each, geom in enumerate(self.j["CityObjects"][theid]["geometry"]):
                    if (
                        self.j["CityObjects"][theid]["geometry"][each]["type"]
                        != "GeometryInstance"
                    ):
                        self.j["CityObjects"][theid]["geometry"][each]["lod"] = str(
                            self.j["CityObjects"][theid]["geometry"][each]["lod"]
                        )
        if "geometry-templates" in self.j:
            for i, g in enumerate(self.j["geometry-templates"]["templates"]):
                self.j["geometry-templates"]["templates"][i]["lod"] = str(
                    self.j["geometry-templates"]["templates"][i]["lod"]
                )
        # -- CityObjectGroup
        # members -> children
        # add parents to children
        for theid in self.j["CityObjects"]:
            if self.j["CityObjects"][theid]["type"] == "CityObjectGroup":
                self.j["CityObjects"][theid]["children"] = self.j["CityObjects"][theid][
                    "members"
                ]
                del self.j["CityObjects"][theid]["members"]
                for ch in self.j["CityObjects"][theid]["children"]:
                    if "parents" not in self.j["CityObjects"][ch]:
                        self.j["CityObjects"][ch]["parents"] = []
                    if theid not in self.j["CityObjects"][ch]["parents"]:
                        self.j["CityObjects"][ch]["parents"].append(theid)
        # -- empty geometries
        for theid in self.j["CityObjects"]:
            if ("geometry" in self.j["CityObjects"][theid]) and (
                len(self.j["CityObjects"][theid]["geometry"]) == 0
            ):
                del self.j["CityObjects"][theid]["geometry"]
        # -- BridgeConstructionElement -> BridgeConstructiveElement
        for theid in self.j["CityObjects"]:
            if self.j["CityObjects"][theid]["type"] == "BridgeConstructionElement":
                self.j["CityObjects"][theid]["type"] = "BridgeConstructiveElement"
        # -- CRS: use the new OGC scheme
        if "metadata" in self.j and "referenceSystem" in self.j["metadata"]:
            s = self.j["metadata"]["referenceSystem"]
            if "epsg" in s.lower():
                self.j["metadata"]["referenceSystem"] = (
                    "https://www.opengis.net/def/crs/EPSG/0/%d"
                    % int(s[s.find("::") + 2 :])
                )
        # -- addresses are now arrays TODO
        for theid in self.j["CityObjects"]:
            if "address" in self.j["CityObjects"][theid]:
                self.j["CityObjects"][theid]["address"] = [
                    self.j["CityObjects"][theid]["address"]
                ]
        # -- metadata calculate
        if "metadata" in self.j:
            v11_properties = {
                "citymodelIdentifier": "identifier",
                "datasetPointOfContact": "pointOfContact",
                "datasetTitle": "title",
                "datasetReferenceDate": "referenceDate",
                "geographicalExtent": "geographicalExtent",
                "referenceSystem": "referenceSystem",
            }
            to_delete = []
            for each in self.j["metadata"]:
                if each not in v11_properties:
                    to_delete.append(each)
            for each in to_delete:
                del self.j["metadata"][each]
            # -- rename to the names
            for each in v11_properties:
                if each in self.j["metadata"]:
                    tmp = self.j["metadata"][each]
                    self.j["metadata"].pop(each)
                    self.j["metadata"][v11_properties[each]] = tmp
        # -- GenericCityObject is no longer, add the Extension GenericCityObject
        gco = False
        for theid in self.j["CityObjects"]:
            if self.j["CityObjects"][theid]["type"] == "GenericCityObject":
                self.j["CityObjects"][theid]["type"] = "+GenericCityObject"
                gco = True
        if gco:
            reasons = (
                '"GenericCityObject" is no longer in v1.1, instead Extensions are used.'
            )
            reasons += (
                ' Your "GenericCityObject" have been changed to "+GenericCityObject"'
            )
            reasons += ' and the simple Extension "Generic" is used.'
            if "extensions" not in self.j:
                self.j["extensions"] = {}
            self.j["extensions"]["Generic"] = {}
            # -- TODO: change URL for Generic Extension
            self.j["extensions"]["Generic"]["url"] = (
                "https://cityjson.org/extensions/download/generic.ext.json"
            )
            self.j["extensions"]["Generic"]["version"] = "1.0"
            return (True, reasons)
        else:
            return (True, "")

    def upgrade_version_v11_v20(self, reasons, digit):
        # -- version
        self.j["version"] = "2.0"
        # -- address in metadata is now a json object
        if (
            "metadata" in self.j
            and "pointOfContact" in self.j["metadata"]
            and "address" in self.j["metadata"]["pointOfContact"]
        ):
            s = self.j["metadata"]["pointOfContact"]["address"]
            self.j["metadata"]["pointOfContact"]["address"] = {}
            self.j["metadata"]["pointOfContact"]["address"]["address"] = s
        # -- GenericCityObject is back!
        if (
            "extensions" in self.j
            and "Generic" in self.j["extensions"]
            and self.j["extensions"]["Generic"]["url"].__contains__(
                "cityjson.org/extensions/download/generic.ext.json"
            )
        ):
            # -- remove the +
            for theid in self.j["CityObjects"]:
                if self.j["CityObjects"][theid]["type"] == "+GenericCityObject":
                    self.j["CityObjects"][theid]["type"] = "GenericCityObject"
            # -- delete the Generic extension
            del self.j["extensions"]["Generic"]
            # -- explain to user
            reasons = '"GenericCityObject" is in v2.0'
            reasons += (
                ' Your "+GenericCityObject" have been changed to "GenericCityObject"'
            )
            return (True, reasons)
        else:
            return (True, "")

    def upgrade_version(self, newversion, digit):
        re = True
        reasons = ""
        if CITYJSON_VERSIONS_SUPPORTED.count(newversion) == 0:
            return (False, "This version is not supported")
        # -- from v0.6
        if self.get_version() == CITYJSON_VERSIONS_SUPPORTED[0]:
            self.upgrade_version_v06_v08()
        # -- v0.8
        if self.get_version() == CITYJSON_VERSIONS_SUPPORTED[1]:
            (re, reasons) = self.upgrade_version_v08_v09(reasons)
        # -- v0.9
        if self.get_version() == CITYJSON_VERSIONS_SUPPORTED[2]:
            (re, reasons) = self.upgrade_version_v09_v10(reasons)
        # -- v1.0 => v1.1
        if self.get_version() == CITYJSON_VERSIONS_SUPPORTED[3]:
            (re, reasons) = self.upgrade_version_v10_v11(reasons, digit)
        # -- v1.1 => v2.0
        if self.get_version() == CITYJSON_VERSIONS_SUPPORTED[4]:
            (re, reasons) = self.upgrade_version_v11_v20(reasons, digit)
        return (re, reasons)

    def export2b3dm(self):
        glb = convert.to_glb(self)
        b3dm = convert.to_b3dm(self, glb)
        return b3dm

    def export2glb(self, do_triangulate=True):
        self.decompress()
        glb = convert.to_glb(self, do_triangulate=do_triangulate)
        return glb

    def export2jsonl(self):
        out = StringIO()
        out.write(self.cityjson_for_features() + "\n")
        # -- take each IDs and create on CityJSONFeature
        for feature in self.generate_features():
            out.write(json.dumps(feature.j, separators=(",", ":")) + "\n")
        return out

    def cityjson_for_features(self):
        """Export a CityJSON object string from the city model.
        The CityJSON object string is meant to be used as the 'first object' in a
        CityJSONFeature stream, thus the 'vertices' and 'CityObjects' are empty."""
        j2 = {}
        j2["type"] = "CityJSON"
        j2["version"] = CITYJSON_VERSIONS_SUPPORTED[-1]
        j2["CityObjects"] = {}
        j2["vertices"] = []
        j2["transform"] = self.j["transform"]
        if "metadata" in self.j:
            j2["metadata"] = self.j["metadata"]
        if "geometry-templates" in self.j:
            j2["geometry-templates"] = self.j["geometry-templates"]
        if "extensions" in self.j:
            j2["extensions"] = self.j["extensions"]
        return json.dumps(j2, separators=(",", ":"))

    def generate_features(self):
        """Generates CityJSONFeatures from the city model.
        Does not output a first CityJSON object. To create the first CityJSON object,
        use :py:func:`cityjson_for_features`.

        Unlike :py:func:`cityjson_for_features`, this function returns the
        CityJSONFeature object (and not a JSON string). This is because downstream
        applications often want to process the CityJSONFeature object further.

        Returns a generator over the CityJSONFeatures.
        """
        idsdone = set()
        theallowedproperties = ["type", "id", "CityObjects", "vertices", "appearance"]
        for theid in self.j["CityObjects"]:
            if ("parents" not in self.j["CityObjects"][theid]) and (
                theid not in idsdone
            ):
                cm2 = self.get_subset_ids([theid])
                cm2.j["type"] = "CityJSONFeature"
                cm2.j["id"] = theid
                # allp = cm2.j
                todelete = []
                for p in cm2.j:
                    if p not in theallowedproperties:
                        todelete.append(p)
                for p in todelete:
                    del cm2.j[p]
                yield cm2
                for theid2 in cm2.j["CityObjects"]:
                    idsdone.add(theid)

    def export2obj(self, sloppy, mtl_fname=None):
        """Exports the city model to a Wavefront OBJ file. If the model has textures and `mtl_fname` is not None, a MTL
        file is also created. The obj file will refer to the .mtl file by the name `mtl_fname`.

        Both the `.obj` and the optional `.mtl` files are returned as StringIO objects.
        """
        imp_digits = math.ceil(abs(math.log(self.j["transform"]["scale"][0], 10)))
        ids = "." + str(imp_digits) + "f"
        self.decompress()
        out = StringIO()
        # -- handle textures
        out_mtl = None
        has_textures = "appearance" in self.j and "textures" in self.j["appearance"]
        export_textures = (
            has_textures and mtl_fname is not None
        )  # if mtl_fname is None, we don't export textures -> for stdout output
        if export_textures:
            out_mtl = StringIO()
            mtl_name = mtl_fname
            out.write("mtllib " + mtl_name + "\n")
            # Create .mtl file
            textures = self.j["appearance"]["textures"]
            for t in textures:
                t["name"] = Path(t["image"]).stem
                out_mtl.write("newmtl {}\n".format(t["name"]))
                out_mtl.write("Ka 1.000 1.000 1.000\n")
                out_mtl.write("Kd 1.000 1.000 1.000\n")
                out_mtl.write("Ks 0.000 0.000 0.000\n")
                out_mtl.write("d 1.0\n")
                out_mtl.write("illum 2\n")
                out_mtl.write("map_Kd {}\n".format(t["image"]))
                out_mtl.write("\n")
        # -- write vertices
        for v in self.j["vertices"]:
            s = format(
                "v {} {} {}\n".format(
                    format(v[0], ids), format(v[1], ids), format(v[2], ids)
                )
            )
            out.write(s)
        vnp = np.array(self.j["vertices"])
        # -- translate to minx,miny
        minx = 9e9
        miny = 9e9
        for each in vnp:
            if each[0] < minx:
                minx = each[0]
            if each[1] < miny:
                miny = each[1]
        for each in vnp:
            each[0] -= minx
            each[1] -= miny

        # -- write texture vertices
        if (
            export_textures
            and "appearance" in self.j
            and "vertices-texture" in self.j["appearance"]
        ):
            for v in self.j["appearance"]["vertices-texture"]:
                s = format("vt {} {}\n".format(format(v[0], ids), format(v[1], ids)))
                out.write(s)
        # -- start with the CO
        for theid in self.j["CityObjects"]:
            if "geometry" not in self.j["CityObjects"][theid]:
                continue
            for geom in self.j["CityObjects"][theid]["geometry"]:
                out.write("o " + str(theid) + "\n")
                if export_textures and "texture" in geom and len(geom["texture"]) > 0:
                    theme = list(geom["texture"].keys())[
                        0
                    ]  # Use the first theme in the dict
                    texture_values = geom["texture"][theme]["values"]
                    if (geom["type"] == "MultiSurface") or (
                        geom["type"] == "CompositeSurface"
                    ):
                        convert.faces_to_obj(
                            geom["boundaries"],
                            out,
                            sloppy,
                            vnp,
                            texture_values,
                            textures,
                        )
                    elif (
                        geom["type"] == "Solid"
                    ):  # depth of geom['texture'] will be one more than for MultiSurface
                        for shell, tex in zip(geom["boundaries"], texture_values):
                            convert.faces_to_obj(shell, out, sloppy, vnp, tex, textures)
                else:
                    if (geom["type"] == "MultiSurface") or (
                        geom["type"] == "CompositeSurface"
                    ):
                        convert.faces_to_obj(geom["boundaries"], out, sloppy, vnp)
                    elif geom["type"] == "Solid":
                        for shell in geom["boundaries"]:
                            convert.faces_to_obj(shell, out, sloppy, vnp)

        self.compress(imp_digits)
        if export_textures:
            return out, out_mtl
        return out

    def export2stl(self, sloppy):
        # TODO: refectoring, duplicated code from 2obj()
        out = StringIO()
        out.write("solid\n")

        # -- translate to minx,miny
        vnp = np.array(self.j["vertices"])
        minx = 9e9
        miny = 9e9
        for each in vnp:
            if each[0] < minx:
                minx = each[0]
            if each[1] < miny:
                miny = each[1]
        for each in vnp:
            each[0] -= minx
            each[1] -= miny

        # -- start with the CO
        for theid in self.j["CityObjects"]:
            for geom in self.j["CityObjects"][theid]["geometry"]:
                if (geom["type"] == "MultiSurface") or (
                    geom["type"] == "CompositeSurface"
                ):
                    for face in geom["boundaries"]:
                        re, b = geom_help.triangulate_face(face, vnp, sloppy)
                        n, bb = geom_help.get_normal_newell(face)
                        if b:
                            for t in re:
                                out.write(
                                    "facet normal %f %f %f\nouter loop\n"
                                    % (n[0], n[1], n[2])
                                )
                                out.write(
                                    "vertex %s %s %s\n"
                                    % (
                                        str(self.j["vertices"][t[0]][0]),
                                        str(self.j["vertices"][t[0]][1]),
                                        str(self.j["vertices"][t[0]][2]),
                                    )
                                )
                                out.write(
                                    "vertex %s %s %s\n"
                                    % (
                                        str(self.j["vertices"][t[1]][0]),
                                        str(self.j["vertices"][t[1]][1]),
                                        str(self.j["vertices"][t[1]][2]),
                                    )
                                )
                                out.write(
                                    "vertex %s %s %s\n"
                                    % (
                                        str(self.j["vertices"][t[2]][0]),
                                        str(self.j["vertices"][t[2]][1]),
                                        str(self.j["vertices"][t[2]][2]),
                                    )
                                )
                                out.write("endloop\nendfacet\n")
                elif geom["type"] == "Solid":
                    for shell in geom["boundaries"]:
                        for i, face in enumerate(shell):
                            re, b = geom_help.triangulate_face(face, vnp, sloppy)
                            n, bb = geom_help.get_normal_newell(face)
                            if b:
                                for t in re:
                                    out.write(
                                        "facet normal %f %f %f\nouter loop\n"
                                        % (n[0], n[1], n[2])
                                    )
                                    out.write(
                                        "vertex %s %s %s\n"
                                        % (
                                            str(self.j["vertices"][t[0]][0]),
                                            str(self.j["vertices"][t[0]][1]),
                                            str(self.j["vertices"][t[0]][2]),
                                        )
                                    )
                                    out.write(
                                        "vertex %s %s %s\n"
                                        % (
                                            str(self.j["vertices"][t[1]][0]),
                                            str(self.j["vertices"][t[1]][1]),
                                            str(self.j["vertices"][t[1]][2]),
                                        )
                                    )
                                    out.write(
                                        "vertex %s %s %s\n"
                                        % (
                                            str(self.j["vertices"][t[2]][0]),
                                            str(self.j["vertices"][t[2]][1]),
                                            str(self.j["vertices"][t[2]][2]),
                                        )
                                    )
                                    out.write("endloop\nendfacet\n")
        out.write("endsolid")
        return out

    def reproject(self, epsg, digit=None):
        """
        Project from one CRS to another.
        First the vertices are decompressed and then they are recompressed.
        Previously we used to recompress by using the same number of digits
        as in the original file. This does not work when switching from projected
        to geographic and vise verse. When the reprojection is such, the
        important digits are set based on the type of CRS.
        """
        if not MODULE_PYPROJ_AVAILABLE:
            raise ModuleNotFoundError(
                """Module 'pyproj' is not available.
                Please install it from https://pypi.org/project/pyproj/"""
            )

        crs_in = CRS(f"EPSG:{self.get_epsg():d}").to_3d()
        crs_out = CRS(f"EPSG:{epsg:d}").to_3d()

        if digit is None:
            if crs_in.is_projected and crs_out.is_geographic:
                imp_digits = 8
            elif crs_in.is_geographic and crs_out.is_projected:
                imp_digits = 3
            else:
                imp_digits = math.ceil(
                    abs(math.log(self.j["transform"]["scale"][0], 10))
                )
                print(imp_digits)
        else:
            imp_digits = digit
        self.decompress()
        # Using TransformerGroup instead of Transformer, because we cannot retrieve the
        # transformer defintion from it.
        # See https://github.com/pyproj4/pyproj/issues/753#issuecomment-737249093
        tg = TransformerGroup(crs_in, crs_out, always_xy=False)
        tg.download_grids(verbose=True)

        with progressbar(self.j["vertices"]) as vertices:
            for v in vertices:
                x, y, z = tg.transformers[0].transform(v[0], v[1], v[2])
                v[0] = x
                v[1] = y
                v[2] = z
        self.set_epsg(epsg)
        self.update_bbox()
        self.update_bbox_each_cityobjects(False)
        self.compress(imp_digits)

    def remove_attribute(self, attr):
        for co in self.j["CityObjects"]:
            if "attributes" in self.j["CityObjects"][co]:
                if attr in self.j["CityObjects"][co]["attributes"]:
                    del self.j["CityObjects"][co]["attributes"][attr]

    def extract_lod(self, thelod):
        def lod_to_string(lod):
            if lod is None:
                return None
            elif isinstance(lod, float):
                return str(round(lod, 1))
            elif isinstance(lod, int):
                return str(lod)
            elif isinstance(lod, str):
                return lod
            else:
                raise ValueError(f"Type {type(lod)} is not allowed as input")

    def rename_attribute(self, oldattr, newattr):
        for co in self.j["CityObjects"]:
            if "attributes" in self.j["CityObjects"][co]:
                if oldattr in self.j["CityObjects"][co]["attributes"]:
                    tmp = self.j["CityObjects"][co]["attributes"][oldattr]
                    self.j["CityObjects"][co]["attributes"][newattr] = tmp
                    del self.j["CityObjects"][co]["attributes"][oldattr]

    def filter_lod(self, thelod):
        for co in self.j["CityObjects"]:
            re = []
            if "geometry" in self.j["CityObjects"][co]:
                for i, g in enumerate(self.j["CityObjects"][co]["geometry"]):
                    if str(g["lod"]) != thelod:
                        re.append(g)
                for each in re:
                    self.j["CityObjects"][co]["geometry"].remove(each)
        self.remove_duplicate_vertices()
        self.remove_orphan_vertices()
        self.update_bbox()

    def translate(self, minxyz: list = None):
        if minxyz is None:
            bbox = self.get_bbox()
            self.j["transform"]["translate"][0] -= bbox[0]
            self.j["transform"]["translate"][1] -= bbox[1]
            self.j["transform"]["translate"][2] -= bbox[2]
        else:
            bbox = minxyz
            self.j["transform"]["translate"][0] += minxyz[0]
            self.j["transform"]["translate"][1] += minxyz[1]
            self.j["transform"]["translate"][2] += minxyz[2]
        self.set_epsg(None)
        self.update_bbox()
        return bbox

    def has_metadata(self):
        """
        Returns whether metadata exist in this CityJSON file or not
        """
        return "metadata" in self.j

    def get_metadata(self):
        """
        Returns the "metadata" property of this CityJSON file

        Raises a KeyError exception if metadata is missing
        """
        if "metadata" not in self.j:
            raise KeyError("Metadata is missing")
        return self.j["metadata"]

    def metadata_extended_remove(self):
        """
        Remove the +metadata-extended in this CityJSON file (if present)
        """
        if "+metadata-extended" in self.j:
            del self.j["+metadata-extended"]
        if "extensions" in self.j and "MetadataExtended" in self.j["extensions"]:
            del self.j["extensions"]["MetadataExtended"]

    def triangulate(self, sloppy):
        """Triangulate the CityJSON file face by face together with the texture information.

        :param sloppy: A boolean, True=mapbox-earcut False=Shewchuk-robust
        """
        vnp = np.array(self.j["vertices"])
        for theid in self.j["CityObjects"]:
            if "geometry" not in self.j["CityObjects"][theid]:
                continue
            for geom in self.j["CityObjects"][theid]["geometry"]:
                sflag = False
                mflag = False
                tflag = False
                slist = []
                mlist = []
                texlist = []
                material = {}
                texture = {}
                # 1.semantics
                if "semantics" in geom:
                    sflag = True
                # 2. materials
                if "material" in geom:
                    material = geom["material"]
                    for item in material.items():
                        if list(item[1].keys())[0] == "value":
                            mlist.append(item[1]["value"])
                        elif list(item[1].keys())[0] == "values":
                            mlist.append([])
                    mflag = True
                # 3. texture
                if "texture" in geom:
                    texture = geom["texture"]
                    tkeys = []
                    for item in texture.items():
                        texlist.append([])
                    for key in texture.keys():
                        tkeys.append(key)
                    tflag = True

                # triangulate the geometry type MultiSurface and CompositeSurface
                if (geom["type"] == "MultiSurface") or (
                    geom["type"] == "CompositeSurface"
                ):
                    tlist1 = []
                    for i, face in enumerate(geom["boundaries"]):
                        tposition = 0
                        tmaplist = []
                        if tflag:
                            for key in tkeys:
                                tmap = {}
                                for ii, f in enumerate(face):
                                    for iii, ff in enumerate(f):
                                        if texture[key]["values"][i][ii][0] is None:
                                            break
                                        else:
                                            tposition = texture[key]["values"][i][ii][0]
                                            tmap[ff] = texture[key]["values"][i][ii][
                                                iii + 1
                                            ]
                                tmaplist.append(tmap)
                        if (len(face) == 1) and (len(face[0]) == 3):
                            re = np.array(face)
                            b = True
                        else:
                            re, b = geom_help.triangulate_face(face, vnp, sloppy)
                            # re, b = geom_help.triangulate_face(face, vnp)

                        if b:
                            for t in re:
                                tlist2 = []
                                tlist2.append(t.tolist())
                                tlist1.append(tlist2)

                                if sflag:
                                    if geom["semantics"]["values"] is None:
                                        slist = None
                                        break
                                    else:
                                        slist.append(geom["semantics"]["values"][i])

                                if mflag:
                                    for j, ls in enumerate(mlist):
                                        if type(ls).__name__ == "list":
                                            ls.append(
                                                material[list(material.keys())[j]][
                                                    "values"
                                                ][i]
                                            )
                                        else:
                                            continue

                                if tflag:
                                    for jj in range(len(tmaplist)):
                                        texlist1 = []
                                        texlist2 = [tposition]

                                        if bool(tmaplist[jj]):
                                            for vindex in t:
                                                texlist2.append(tmaplist[jj][vindex])
                                        else:
                                            texlist2[0] = None
                                        texlist1.append(texlist2)
                                        texlist[jj].append(texlist1)

                    geom["boundaries"] = tlist1
                    if sflag:
                        geom["semantics"]["values"] = slist

                    if mflag:
                        for j, item in enumerate(material.items()):
                            item[1][list(item[1].keys())[0]] = mlist[j]
                        geom["material"] = material

                    if tflag:
                        for j, item in enumerate(texture.items()):
                            item[1][list(item[1].keys())[0]] = texlist[j]
                        geom["texture"] = texture

                # triangulate the geometry type Solid
                elif geom["type"] == "Solid":
                    tlist1 = []
                    minit = copy.deepcopy(mlist)
                    texinit = copy.deepcopy(texlist)
                    for sidx, shell in enumerate(geom["boundaries"]):
                        slist1 = []
                        tlist2 = []
                        texlist0 = copy.deepcopy(texinit)
                        mlist1 = copy.deepcopy(minit)
                        for i, face in enumerate(shell):
                            tposition = 0
                            tmaplist = []
                            if tflag:
                                for key in tkeys:
                                    tmap = {}
                                    for ii, f in enumerate(face):
                                        for iii, ff in enumerate(f):
                                            if (
                                                texture[key]["values"][sidx][i][ii][0]
                                                is None
                                            ):
                                                break
                                            else:
                                                tposition = texture[key]["values"][
                                                    sidx
                                                ][i][ii][0]
                                                tmap[ff] = texture[key]["values"][sidx][
                                                    i
                                                ][ii][iii + 1]
                                    tmaplist.append(tmap)
                            if (len(face) == 1) and (len(face[0]) == 3):
                                re = np.array(face)
                                b = True
                            else:
                                re, b = geom_help.triangulate_face(face, vnp, sloppy)
                            if b:
                                for t in re:
                                    tlist3 = []
                                    tlist3.append(t.tolist())
                                    tlist2.append(tlist3)
                                    if sflag:
                                        if geom["semantics"]["values"] is None:
                                            slist = None
                                            break
                                        elif geom["semantics"]["values"][sidx] is None:
                                            slist1 = None
                                            break
                                        else:
                                            slist1.append(
                                                geom["semantics"]["values"][sidx][i]
                                            )
                                    if mflag:
                                        for j, ls in enumerate(mlist1):
                                            if type(ls).__name__ == "list":
                                                ls.append(
                                                    material[list(material.keys())[j]][
                                                        "values"
                                                    ][0][i]
                                                )
                                            else:
                                                continue

                                    if tflag:
                                        for jj in range(len(tmaplist)):
                                            texlist1 = []
                                            texlist2 = [tposition]

                                            if bool(tmaplist[jj]):
                                                for vindex in t:
                                                    texlist2.append(
                                                        tmaplist[jj][vindex]
                                                    )
                                            else:
                                                texlist2[0] = None
                                            texlist1.append(texlist2)
                                            texlist0[jj].append(texlist1)

                        tlist1.append(tlist2)
                        if slist is not None:
                            slist.append(slist1)
                        for j, ls in enumerate(mlist):
                            if type(ls).__name__ == "list" and len(mlist1[j]) != 0:
                                ls.append(mlist1[j])
                            else:
                                continue
                        for j, ls in enumerate(texlist):
                            ls.append(texlist0[j])

                    geom["boundaries"] = tlist1
                    if sflag:
                        geom["semantics"]["values"] = slist

                    if mflag:
                        for j, item in enumerate(material.items()):
                            item[1][list(item[1].keys())[0]] = mlist[j]
                        geom["material"] = material

                    if tflag:
                        for j, item in enumerate(texture.items()):
                            item[1][list(item[1].keys())[0]] = texlist[j]
                        geom["texture"] = texture

                # triangulate the geometry type MultiSolid and CompositeSolid
                elif (geom["type"] == "MultiSolid") or (
                    geom["type"] == "CompositeSolid"
                ):
                    tlist1 = []
                    minit = copy.deepcopy(mlist)
                    texinit = copy.deepcopy(texlist)
                    for solididx, solid in enumerate(geom["boundaries"]):
                        slist1 = []
                        tlist2 = []
                        mlist1 = copy.deepcopy(minit)
                        texlist0 = copy.deepcopy(texinit)
                        for sidx, shell in enumerate(solid):
                            slist2 = []
                            tlist3 = []
                            mlist2 = copy.deepcopy(minit)
                            texlist1 = copy.deepcopy(texinit)
                            for i, face in enumerate(shell):
                                tposition = 0
                                tmaplist = []
                                if tflag:
                                    for key in tkeys:
                                        tmap = {}
                                        for ii, f in enumerate(face):
                                            for iii, ff in enumerate(f):
                                                if (
                                                    texture[key]["values"][solididx][
                                                        sidx
                                                    ][i][ii][0]
                                                    is None
                                                ):
                                                    break
                                                else:
                                                    tposition = texture[key]["values"][
                                                        solididx
                                                    ][sidx][i][ii][0]
                                                    tmap[ff] = texture[key]["values"][
                                                        solididx
                                                    ][sidx][i][ii][iii + 1]
                                        tmaplist.append(tmap)
                                if (len(face) == 1) and (len(face[0]) == 3):
                                    re = np.array(face)
                                    b = True
                                else:
                                    re, b = geom_help.triangulate_face(
                                        face, vnp, sloppy
                                    )
                                if b:
                                    for t in re:
                                        tlist4 = []
                                        tlist4.append(t.tolist())
                                        tlist3.append(tlist4)

                                        if sflag:
                                            if geom["semantics"]["values"] is None:
                                                slist = None
                                                break
                                            if (
                                                geom["semantics"]["values"][solididx]
                                                is None
                                            ):
                                                slist1 = None
                                                break
                                            elif (
                                                geom["semantics"]["values"][solididx][
                                                    sidx
                                                ]
                                                is None
                                            ):
                                                slist2 = None
                                                break
                                            else:
                                                slist2.append(
                                                    geom["semantics"]["values"][
                                                        solididx
                                                    ][sidx][i]
                                                )

                                        if mflag:
                                            for j, ls in enumerate(mlist2):
                                                if type(ls).__name__ == "list":
                                                    ls.append(
                                                        material[
                                                            list(material.keys())[j]
                                                        ]["values"][0][0][i]
                                                    )
                                                else:
                                                    continue

                                        if tflag:
                                            for jj in range(len(tmaplist)):
                                                texlist2 = []
                                                texlist3 = [tposition]

                                                if bool(tmaplist[jj]):
                                                    for vindex in t:
                                                        texlist3.append(
                                                            tmaplist[jj][vindex]
                                                        )
                                                else:
                                                    texlist3[0] = None
                                                texlist2.append(texlist3)
                                                texlist1[jj].append(texlist2)

                            tlist2.append(tlist3)
                            if slist1 is not None:
                                slist1.append(slist2)
                            for j, ls in enumerate(mlist1):
                                if type(ls).__name__ == "list" and len(mlist2[j]) != 0:
                                    ls.append(mlist2[j])
                                else:
                                    continue
                            for j, ls in enumerate(texlist0):
                                ls.append(texlist1[j])

                        tlist1.append(tlist2)
                        if slist is not None:
                            slist.append(slist1)
                        for j, ls in enumerate(mlist):
                            if type(ls).__name__ == "list" and len(mlist1[j]) != 0:
                                ls.append(mlist1[j])
                            else:
                                continue
                        for j, ls in enumerate(texlist):
                            ls.append(texlist0[j])

                    geom["boundaries"] = tlist1
                    if sflag:
                        geom["semantics"]["values"] = slist
                    if mflag:
                        for j, item in enumerate(material.items()):
                            item[1][list(item[1].keys())[0]] = mlist[j]
                        geom["material"] = material
                    if tflag:
                        for j, item in enumerate(texture.items()):
                            item[1][list(item[1].keys())[0]] = texlist[j]
                        geom["texture"] = texture

    def is_triangulated(self):
        """
        Check if the CityJSON file is *fully* triangulated. Return true if it's triangulated, return false if it's not.
        """

        for theid in self.j["CityObjects"]:
            if "geometry" in self.j["CityObjects"][theid]:
                for geom in self.j["CityObjects"][theid]["geometry"]:
                    if (geom["type"] == "MultiSurface") or (
                        geom["type"] == "CompositeSurface"
                    ):
                        for face in geom["boundaries"]:
                            for f in face:
                                if len(f) != 3:
                                    return False
                    elif geom["type"] == "Solid":
                        for shell in geom["boundaries"]:
                            for face in shell:
                                for f in face:
                                    if len(f) != 3:
                                        return False
                    elif (geom["type"] == "MultiSolid") or (
                        geom["type"] == "CompositeSolid"
                    ):
                        for solid in geom["boundaries"]:
                            for shell in solid:
                                for face in shell:
                                    for f in face:
                                        if len(f) != 3:
                                            return False
        return True
